#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import uuid
import pytest
import aiohttp
from unittest.mock import MagicMock
from tests.utils import AsyncioMagicMock, asyncio_patch
from unittest.mock import patch
from uuid import uuid4

from gns3server.controller.project import Project
from gns3server.controller.template import Template
from gns3server.controller.ports.ethernet_port import EthernetPort
from gns3server.config import Config


@pytest.fixture
def project(controller):
    return Project(controller=controller, name="Test")


@pytest.fixture
def node(controller, project, async_run):
    compute = MagicMock()
    compute.id = "local"

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    return node


def test_affect_uuid():
    p = Project(name="Test")
    assert len(p.id) == 36

    p = Project(project_id='00010203-0405-0607-0809-0a0b0c0d0e0f', name="Test 2")
    assert p.id == '00010203-0405-0607-0809-0a0b0c0d0e0f'


def test_json(tmpdir):
    p = Project(name="Test")
    assert p.__json__() == {
        "name": "Test",
        "project_id": p.id,
        "path": p.path,
        "status": "opened",
        "filename": "Test.gns3",
        "auto_start": False,
        "auto_close": True,
        "auto_open": False,
        "scene_width": 2000,
        "scene_height": 1000,
        "zoom": 100,
        "show_grid": False,
        "show_interface_labels": False,
        "show_layers": False,
        "snap_to_grid": False,
        "grid_size": 75,
        "drawing_grid_size": 25,
        "supplier": None,
        "variables": None
    }


def test_update(controller, async_run):
    project = Project(controller=controller, name="Hello")
    project.emit_notification = MagicMock()
    assert project.name == "Hello"
    async_run(project.update(name="World"))
    assert project.name == "World"
    project.emit_notification.assert_any_call("project.updated", project.__json__())


def test_update_on_compute(controller, async_run):
    variables = [{"name": "TEST", "value": "VAL1"}]
    compute = MagicMock()
    compute.id = "local"
    project = Project(controller=controller, name="Test")
    project._project_created_on_compute = [compute]
    project.emit_notification = MagicMock()

    async_run(project.update(variables=variables))

    compute.put.assert_any_call('/projects/{}'.format(project.id), {
        "variables": variables
    })


def test_path(tmpdir):

    directory = Config.instance().get_section_config("Server").get("projects_path")

    with patch("gns3server.utils.path.get_default_project_directory", return_value=directory):
        p = Project(project_id=str(uuid4()), name="Test")
        assert p.path == os.path.join(directory, p.id)
        assert os.path.exists(os.path.join(directory, p.id))


def test_path_exist(tmpdir):
    """
    Should raise an error when you try to owerwrite
    an existing project
    """
    os.makedirs(str(tmpdir / "demo"))
    with pytest.raises(aiohttp.web.HTTPForbidden):
        p = Project(name="Test", path=str(tmpdir / "demo"))


def test_init_path(tmpdir):

    p = Project(path=str(tmpdir), project_id=str(uuid4()), name="Test")
    assert p.path == str(tmpdir)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Not supported on Windows")
def test_changing_path_with_quote_not_allowed(tmpdir):
    with pytest.raises(aiohttp.web.HTTPForbidden):
        p = Project(project_id=str(uuid4()), name="Test")
        p.path = str(tmpdir / "project\"53")


def test_captures_directory(tmpdir):
    p = Project(path=str(tmpdir / "capturestest"), name="Test")
    assert p.captures_directory == str(tmpdir / "capturestest" / "project-files" / "captures")
    assert os.path.exists(p.captures_directory)


def test_add_node_local(async_run, controller):
    """
    For a local server we send the project path
    """
    compute = MagicMock()
    compute.id = "local"
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_script": "test.cfg"}))
    assert node.id in project._nodes

    compute.post.assert_any_call('/projects', data={
        "name": project._name,
        "project_id": project._id,
        "path": project._path,
    })
    compute.post.assert_any_call('/projects/{}/vpcs/nodes'.format(project.id),
                                 data={'node_id': node.id,
                                       'startup_script': 'test.cfg',
                                       'name': 'test'},
                                 timeout=1200)
    assert compute in project._project_created_on_compute
    project.emit_notification.assert_any_call("node.created", node.__json__())


def test_add_node_non_local(async_run, controller):
    """
    For a non local server we do not send the project path
    """
    compute = MagicMock()
    compute.id = "remote"
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_script": "test.cfg"}))

    compute.post.assert_any_call('/projects', data={
        "name": project._name,
        "project_id": project._id
    })
    compute.post.assert_any_call('/projects/{}/vpcs/nodes'.format(project.id),
                                 data={'node_id': node.id,
                                       'startup_script': 'test.cfg',
                                       'name': 'test'},
                                 timeout=1200)
    assert compute in project._project_created_on_compute
    project.emit_notification.assert_any_call("node.created", node.__json__())


def test_add_node_from_template(async_run, controller):
    """
    For a local server we send the project path
    """
    compute = MagicMock()
    compute.id = "local"
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()
    template = Template(str(uuid.uuid4()), {
        "compute_id": "local",
        "name": "Test",
        "template_type": "vpcs",
        "builtin": False,
    })
    controller.template_manager.templates[template.id] = template
    controller._computes["local"] = compute

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node_from_template(template.id, x=23, y=12))
    compute.post.assert_any_call('/projects', data={
        "name": project._name,
        "project_id": project._id,
        "path": project._path
    })

    assert compute in project._project_created_on_compute
    project.emit_notification.assert_any_call("node.created", node.__json__())


def test_add_builtin_node_from_template(async_run, controller):
    """
    For a local server we send the project path
    """
    compute = MagicMock()
    compute.id = "local"
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()
    template = Template(str(uuid.uuid4()), {
        "name": "Builtin-switch",
        "template_type": "ethernet_switch",
    }, builtin=True)
    controller.template_manager.templates[template.id] = template
    template.__json__()
    controller._computes["local"] = compute

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node_from_template(template.id, x=23, y=12, compute_id="local"))
    compute.post.assert_any_call('/projects', data={
        "name": project._name,
        "project_id": project._id,
        "path": project._path
    })

    assert compute in project._project_created_on_compute
    project.emit_notification.assert_any_call("node.created", node.__json__())


def test_delete_node(async_run, controller):
    """
    For a local server we send the project path
    """
    compute = MagicMock()
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.id in project._nodes
    async_run(project.delete_node(node.id))
    assert node.id not in project._nodes

    compute.delete.assert_any_call('/projects/{}/vpcs/nodes/{}'.format(project.id, node.id))
    project.emit_notification.assert_any_call("node.deleted", node.__json__())


def test_delete_locked_node(async_run, controller):
    """
    For a local server we send the project path
    """
    compute = MagicMock()
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.id in project._nodes
    node.locked = True
    with pytest.raises(aiohttp.web_exceptions.HTTPConflict):
        async_run(project.delete_node(node.id))


def test_delete_node_delete_link(async_run, controller):
    """
    Delete a node delete all the node connected
    """
    compute = MagicMock()
    project = Project(controller=controller, name="Test")
    project.emit_notification = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))

    link = async_run(project.add_link())
    async_run(link.add_node(node, 0, 0))

    async_run(project.delete_node(node.id))
    assert node.id not in project._nodes
    assert link.id not in project._links

    compute.delete.assert_any_call('/projects/{}/vpcs/nodes/{}'.format(project.id, node.id))
    project.emit_notification.assert_any_call("node.deleted", node.__json__())
    project.emit_notification.assert_any_call("link.deleted", link.__json__())


def test_get_node(async_run, controller):
    compute = MagicMock()
    project = Project(controller=controller, name="Test")

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    vm = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert project.get_node(vm.id) == vm

    with pytest.raises(aiohttp.web_exceptions.HTTPNotFound):
        project.get_node("test")

    # Raise an error if the project is not opened
    async_run(project.close())
    with pytest.raises(aiohttp.web.HTTPForbidden):
        project.get_node(vm.id)

def test_list_nodes(async_run, controller):
    compute = MagicMock()
    project = Project(controller=controller, name="Test")

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    vm = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert len(project.nodes) == 1
    assert isinstance(project.nodes, dict)

    async_run(project.close())
    assert len(project.nodes) == 1
    assert isinstance(project.nodes, dict)


def test_add_link(async_run, project, controller):
    compute = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    vm1 = async_run(project.add_node(compute, "test1", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    vm1._ports = [EthernetPort("E0", 0, 3, 1)]
    vm2 = async_run(project.add_node(compute, "test2", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    vm2._ports = [EthernetPort("E0", 0, 4, 2)]
    project.emit_notification = MagicMock()
    link = async_run(project.add_link())
    async_run(link.add_node(vm1, 3, 1))
    with asyncio_patch("gns3server.controller.udp_link.UDPLink.create") as mock_udp_create:
        async_run(link.add_node(vm2, 4, 2))
    assert mock_udp_create.called
    assert len(link._nodes) == 2
    project.emit_notification.assert_any_call("link.created", link.__json__())


def test_list_links(async_run, project):
    compute = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    link = async_run(project.add_link())
    assert len(project.links) == 1

    async_run(project.close())
    assert len(project.links) == 1


def test_get_link(async_run, project):
    compute = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    link = async_run(project.add_link())
    assert project.get_link(link.id) == link

    with pytest.raises(aiohttp.web_exceptions.HTTPNotFound):
        project.get_link("test")


def test_delete_link(async_run, project, controller):
    compute = MagicMock()

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    assert len(project._links) == 0
    link = async_run(project.add_link())
    assert len(project._links) == 1
    project.emit_notification = MagicMock()
    async_run(project.delete_link(link.id))
    project.emit_notification.assert_any_call("link.deleted", link.__json__())
    assert len(project._links) == 0


def test_add_drawing(async_run, project, controller):
    project.emit_notification = MagicMock()

    drawing = async_run(project.add_drawing(None, svg="<svg></svg>"))
    assert len(project._drawings) == 1
    project.emit_notification.assert_any_call("drawing.created", drawing.__json__())


def test_get_drawing(async_run, project):
    drawing = async_run(project.add_drawing(None))
    assert project.get_drawing(drawing.id) == drawing

    with pytest.raises(aiohttp.web_exceptions.HTTPNotFound):
        project.get_drawing("test")


def test_list_drawing(async_run, project):
    drawing = async_run(project.add_drawing(None))
    assert len(project.drawings) == 1

    async_run(project.close())
    assert len(project.drawings) == 1


def test_delete_drawing(async_run, project, controller):
    assert len(project._drawings) == 0
    drawing = async_run(project.add_drawing())
    assert len(project._drawings) == 1
    project.emit_notification = MagicMock()
    async_run(project.delete_drawing(drawing.id))
    project.emit_notification.assert_any_call("drawing.deleted", drawing.__json__())
    assert len(project._drawings) == 0


def test_clean_pictures(async_run, project, controller):
    """
    When a project is close old pictures should be removed
    """

    drawing = async_run(project.add_drawing())
    drawing._svg = "test.png"
    open(os.path.join(project.pictures_directory, "test.png"), "w+").close()
    open(os.path.join(project.pictures_directory, "test2.png"), "w+").close()
    async_run(project.close())
    assert os.path.exists(os.path.join(project.pictures_directory, "test.png"))
    assert not os.path.exists(os.path.join(project.pictures_directory, "test2.png"))


def test_clean_pictures_and_keep_supplier_logo(async_run, project, controller):
    """
    When a project is close old pictures should be removed
    """
    project.supplier = {
        'logo': 'logo.png'
    }

    drawing = async_run(project.add_drawing())
    drawing._svg = "test.png"
    open(os.path.join(project.pictures_directory, "test.png"), "w+").close()
    open(os.path.join(project.pictures_directory, "test2.png"), "w+").close()
    open(os.path.join(project.pictures_directory, "logo.png"), "w+").close()

    async_run(project.close())
    assert os.path.exists(os.path.join(project.pictures_directory, "test.png"))
    assert not os.path.exists(os.path.join(project.pictures_directory, "test2.png"))
    assert os.path.exists(os.path.join(project.pictures_directory, "logo.png"))


def test_delete(async_run, project, controller):
    assert os.path.exists(project.path)
    async_run(project.delete())
    assert not os.path.exists(project.path)


def test_dump():
    directory = Config.instance().get_section_config("Server").get("projects_path")

    with patch("gns3server.utils.path.get_default_project_directory", return_value=directory):
        p = Project(project_id='00010203-0405-0607-0809-0a0b0c0d0e0f', name="Test")
        p.dump()
        with open(os.path.join(directory, p.id, "Test.gns3")) as f:
            content = f.read()
            assert "00010203-0405-0607-0809-0a0b0c0d0e0f" in content


def test_open_close(async_run, controller):
    project = Project(controller=controller, name="Test")
    assert project.status == "opened"
    async_run(project.close())
    project.start_all = AsyncioMagicMock()
    async_run(project.open())
    assert not project.start_all.called
    assert project.status == "opened"
    project.emit_notification = MagicMock()
    async_run(project.close())
    assert project.status == "closed"
    project.emit_notification.assert_any_call("project.closed", project.__json__())


def test_open_auto_start(async_run, controller):
    project = Project(controller=controller, name="Test", auto_start=True)
    assert project.status == "opened"
    async_run(project.close())
    project.start_all = AsyncioMagicMock()
    async_run(project.open())
    assert project.start_all.called


def test_is_running(project, async_run, node):
    """
    If a node is started or paused return True
    """

    assert project.is_running() is False
    node._status = "started"
    assert project.is_running() is True


def test_duplicate(project, async_run, controller):
    """
    Duplicate a project, the node should remain on the remote server
    if they were on remote server
    """
    compute = MagicMock()
    compute.id = "remote"
    compute.list_files = AsyncioMagicMock(return_value=[])
    controller._computes["remote"] = compute

    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    remote_vpcs = async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))

    # We allow node not allowed for standard import / export
    remote_virtualbox = async_run(project.add_node(compute, "test", None, node_type="vmware", properties={"startup_config": "test.cfg"}))

    new_project = async_run(project.duplicate(name="Hello"))
    assert new_project.id != project.id
    assert new_project.name == "Hello"

    async_run(new_project.open())

    assert list(new_project.nodes.values())[0].compute.id == "remote"
    assert list(new_project.nodes.values())[1].compute.id == "remote"


def test_snapshots(project):
    """
    List the snapshots
    """
    os.makedirs(os.path.join(project.path, "snapshots"))
    open(os.path.join(project.path, "snapshots", "test1_260716_103713.gns3project"), "w+").close()
    project.reset()

    assert len(project.snapshots) == 1
    assert list(project.snapshots.values())[0].name == "test1"


def test_get_snapshot(project):
    os.makedirs(os.path.join(project.path, "snapshots"))
    open(os.path.join(project.path, "snapshots", "test1.gns3project"), "w+").close()
    project.reset()

    snapshot = list(project.snapshots.values())[0]
    assert project.get_snapshot(snapshot.id) == snapshot

    with pytest.raises(aiohttp.web_exceptions.HTTPNotFound):
        project.get_snapshot("BLU")


def test_delete_snapshot(project, async_run):
    os.makedirs(os.path.join(project.path, "snapshots"))
    open(os.path.join(project.path, "snapshots", "test1_260716_103713.gns3project"), "w+").close()
    project.reset()

    snapshot = list(project.snapshots.values())[0]
    assert project.get_snapshot(snapshot.id) == snapshot

    async_run(project.delete_snapshot(snapshot.id))

    with pytest.raises(aiohttp.web_exceptions.HTTPNotFound):
        project.get_snapshot(snapshot.id)

    assert not os.path.exists(os.path.join(project.path, "snapshots", "test1.gns3project"))


def test_snapshot(project, async_run):
    """
    Create a snapshot
    """
    assert len(project.snapshots) == 0

    snapshot = async_run(project.snapshot("test1"))
    assert snapshot.name == "test1"

    assert len(project.snapshots) == 1
    assert list(project.snapshots.values())[0].name == "test1"

    # Raise a conflict if name is already use
    with pytest.raises(aiohttp.web_exceptions.HTTPConflict):
        snapshot = async_run(project.snapshot("test1"))


def test_start_all(project, async_run):
    compute = MagicMock()
    compute.id = "local"
    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    for node_i in range(0, 10):
        async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))

    compute.post = AsyncioMagicMock()
    async_run(project.start_all())
    assert len(compute.post.call_args_list) == 10


def test_stop_all(project, async_run):
    compute = MagicMock()
    compute.id = "local"
    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    for node_i in range(0, 10):
        async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))

    compute.post = AsyncioMagicMock()
    async_run(project.stop_all())
    assert len(compute.post.call_args_list) == 10


def test_suspend_all(project, async_run):
    compute = MagicMock()
    compute.id = "local"
    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    for node_i in range(0, 10):
        async_run(project.add_node(compute, "test", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))

    compute.post = AsyncioMagicMock()
    async_run(project.suspend_all())
    assert len(compute.post.call_args_list) == 10


def test_node_name(project, async_run):
    compute = MagicMock()
    compute.id = "local"
    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    node = async_run(project.add_node(compute, "test-{0}", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "test-1"
    node = async_run(project.add_node(compute, "test-{0}", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "test-2"
    node = async_run(project.add_node(compute, "hello world-{0}", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "helloworld-1"
    node = async_run(project.add_node(compute, "hello world-{0}", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "helloworld-2"
    node = async_run(project.add_node(compute, "VPCS-1", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "VPCS-1"
    node = async_run(project.add_node(compute, "VPCS-1", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "VPCS-2"

    node = async_run(project.add_node(compute, "R3", None, node_type="vpcs", properties={"startup_config": "test.cfg"}))
    assert node.name == "R3"


def test_duplicate_node(project, async_run):
    compute = MagicMock()
    compute.id = "local"
    response = MagicMock()
    response.json = {"console": 2048}
    compute.post = AsyncioMagicMock(return_value=response)

    original = async_run(project.add_node(
        compute,
        "test",
        None,
        node_type="vpcs",
        properties={
            "startup_config": "test.cfg"
        }))
    new_node = async_run(project.duplicate_node(original, 42, 10, 11))
    assert new_node.x == 42



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from gns3server.web.documentation import Documentation
from gns3server.handlers import *
from gns3server.web.route import Route


def test_documentation_write(tmpdir):
    os.makedirs(str(tmpdir / "api/examples"))
    with open(str(tmpdir / "api/examples/compute_post_projectsprojectidvirtualboxnodes.txt"), "w+") as f:
        f.write("curl test")

    Documentation(Route, str(tmpdir)).write()

    assert os.path.exists(str(tmpdir / "api"))
    assert os.path.exists(str(tmpdir / "api" / "v2" / "compute"))
    assert os.path.exists(str(tmpdir / "api" / "v2" / "compute" / "virtualbox.rst"))
    assert os.path.exists(str(tmpdir / "api" / "v2" / "compute" / "virtualbox"))
    assert os.path.exists(str(tmpdir / "api" / "v2" / "compute" / "virtualbox" / "virtualboxvms.rst"))
    with open(str(tmpdir / "api" / "v2" / "compute" / "virtualbox" / "projectsprojectidvirtualboxnodes.rst")) as f:
        content = f.read()
        assert "Sample session" in content
        assert "literalinclude:: ../../../examples/compute_post_projectsprojectidvirtualboxnodes.txt" in content

    assert os.path.exists(str(tmpdir / "api" / "v2" / "controller" / "compute.rst"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
from .template import BASE_TEMPLATE_PROPERTIES
from .port import PORT_OBJECT_SCHEMA


CLOUD_TEMPLATE_PROPERTIES = {
    "ports_mapping": {
        "type": "array",
        "items": [PORT_OBJECT_SCHEMA],
        "default": []
    },
    "remote_console_host": {
        "description": "Remote console host or IP",
        "type": ["string"],
        "minLength": 1,
        "default": "127.0.0.1"
    },
    "remote_console_port": {
        "description": "Console TCP port",
        "minimum": 1,
        "maximum": 65535,
        "type": "integer",
        "default": 23
    },
    "remote_console_type": {
        "description": "Console type",
        "enum": ["telnet", "vnc", "spice", "http", "https", "none"],
        "default": "none"
    },
    "remote_console_http_path": {
        "description": "Path of the remote web interface",
        "type": "string",
        "minLength": 1,
        "default": "/"
    },
}

CLOUD_TEMPLATE_PROPERTIES.update(copy.deepcopy(BASE_TEMPLATE_PROPERTIES))
CLOUD_TEMPLATE_PROPERTIES["category"]["default"] = "guest"
CLOUD_TEMPLATE_PROPERTIES["default_name_format"]["default"] = "Cloud{0}"
CLOUD_TEMPLATE_PROPERTIES["symbol"]["default"] = ":/symbols/cloud.svg"

CLOUD_TEMPLATE_OBJECT_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "A cloud template object",
    "type": "object",
    "properties": CLOUD_TEMPLATE_PROPERTIES,
    "additionalProperties": False
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from aiohttp.web import HTTPConflict
from gns3server.web.route import Route
from gns3server.schemas.nio import NIO_SCHEMA
from gns3server.schemas.node import NODE_CAPTURE_SCHEMA
from gns3server.compute.vpcs import VPCS

from gns3server.schemas.vpcs import (
    VPCS_CREATE_SCHEMA,
    VPCS_UPDATE_SCHEMA,
    VPCS_OBJECT_SCHEMA
)


class VPCSHandler:
    """
    API entry points for VPCS.
    """

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes",
        parameters={
            "project_id": "Project UUID"
        },
        status_codes={
            201: "Instance created",
            400: "Invalid request",
            409: "Conflict"
        },
        description="Create a new VPCS instance",
        input=VPCS_CREATE_SCHEMA,
        output=VPCS_OBJECT_SCHEMA)
    async def create(request, response):

        vpcs = VPCS.instance()
        vm = await vpcs.create_node(request.json["name"],
                                         request.match_info["project_id"],
                                         request.json.get("node_id"),
                                         console=request.json.get("console"),
                                         console_type=request.json.get("console_type", "telnet"),
                                         startup_script=request.json.get("startup_script"))
        response.set_status(201)
        response.json(vm)

    @Route.get(
        r"/projects/{project_id}/vpcs/nodes/{node_id}",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            200: "Success",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Get a VPCS instance",
        output=VPCS_OBJECT_SCHEMA)
    def show(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        response.json(vm)

    @Route.put(
        r"/projects/{project_id}/vpcs/nodes/{node_id}",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            200: "Instance updated",
            400: "Invalid request",
            404: "Instance doesn't exist",
            409: "Conflict"
        },
        description="Update a VPCS instance",
        input=VPCS_UPDATE_SCHEMA,
        output=VPCS_OBJECT_SCHEMA)
    def update(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        vm.name = request.json.get("name", vm.name)
        vm.console = request.json.get("console", vm.console)
        vm.console_type = request.json.get("console_type", vm.console_type)
        vm.updated()
        response.json(vm)

    @Route.delete(
        r"/projects/{project_id}/vpcs/nodes/{node_id}",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            204: "Instance deleted",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Delete a VPCS instance")
    async def delete(request, response):

        await VPCS.instance().delete_node(request.match_info["node_id"])
        response.set_status(204)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/duplicate",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            201: "Instance duplicated",
            404: "Instance doesn't exist"
        },
        description="Duplicate a VPCS instance")
    async def duplicate(request, response):

        new_node = await VPCS.instance().duplicate_node(
            request.match_info["node_id"],
            request.json["destination_node_id"]
        )
        response.set_status(201)
        response.json(new_node)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/start",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            204: "Instance started",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Start a VPCS instance",
        output=VPCS_OBJECT_SCHEMA)
    async def start(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        await vm.start()
        response.json(vm)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/stop",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            204: "Instance stopped",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Stop a VPCS instance")
    async def stop(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        await vm.stop()
        response.set_status(204)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/suspend",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID"
        },
        status_codes={
            204: "Instance suspended",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Suspend a VPCS instance (does nothing)")
    def suspend(request, response):

        vpcs_manager = VPCS.instance()
        vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        response.set_status(204)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/reload",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
        },
        status_codes={
            204: "Instance reloaded",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Reload a VPCS instance")
    async def reload(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        await vm.reload()
        response.set_status(204)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/adapters/{adapter_number:\d+}/ports/{port_number:\d+}/nio",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
            "adapter_number": "Network adapter where the nio is located",
            "port_number": "Port where the nio should be added"
        },
        status_codes={
            201: "NIO created",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Add a NIO to a VPCS instance",
        input=NIO_SCHEMA,
        output=NIO_SCHEMA)
    async def create_nio(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        nio_type = request.json["type"]
        if nio_type not in ("nio_udp", "nio_tap"):
            raise HTTPConflict(text="NIO of type {} is not supported".format(nio_type))
        port_number = int(request.match_info["port_number"])
        nio = vpcs_manager.create_nio(request.json)
        await vm.port_add_nio_binding(port_number, nio)
        response.set_status(201)
        response.json(nio)

    @Route.put(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/adapters/{adapter_number:\d+}/ports/{port_number:\d+}/nio",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
            "adapter_number": "Network adapter where the nio is located",
            "port_number": "Port from where the nio should be updated"
        },
        status_codes={
            201: "NIO updated",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        input=NIO_SCHEMA,
        output=NIO_SCHEMA,
        description="Update a NIO on a VPCS instance")
    async def update_nio(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        port_number = int(request.match_info["port_number"])
        nio = vm.get_nio(port_number)
        if "filters" in request.json:
            nio.filters = request.json["filters"]
        await vm.port_update_nio_binding(port_number, nio)
        response.set_status(201)
        response.json(request.json)

    @Route.delete(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/adapters/{adapter_number:\d+}/ports/{port_number:\d+}/nio",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
            "adapter_number": "Network adapter where the nio is located",
            "port_number": "Port from where the nio should be removed"
        },
        status_codes={
            204: "NIO deleted",
            400: "Invalid request",
            404: "Instance doesn't exist"
        },
        description="Remove a NIO from a VPCS instance")
    async def delete_nio(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        port_number = int(request.match_info["port_number"])
        await vm.port_remove_nio_binding(port_number)
        response.set_status(204)

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/adapters/{adapter_number:\d+}/ports/{port_number:\d+}/start_capture",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
            "adapter_number": "Adapter to start a packet capture",
            "port_number": "Port on the adapter"
        },
        status_codes={
            200: "Capture started",
            400: "Invalid request",
            404: "Instance doesn't exist",
        },
        description="Start a packet capture on a VPCS instance",
        input=NODE_CAPTURE_SCHEMA)
    async def start_capture(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        port_number = int(request.match_info["port_number"])
        pcap_file_path = os.path.join(vm.project.capture_working_directory(), request.json["capture_file_name"])
        await vm.start_capture(port_number, pcap_file_path)
        response.json({"pcap_file_path": pcap_file_path})

    @Route.post(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/adapters/{adapter_number:\d+}/ports/{port_number:\d+}/stop_capture",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
            "adapter_number": "Adapter to stop a packet capture",
            "port_number": "Port on the adapter"
        },
        status_codes={
            204: "Capture stopped",
            400: "Invalid request",
            404: "Instance doesn't exist",
        },
        description="Stop a packet capture on a VPCS instance")
    async def stop_capture(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        port_number = int(request.match_info["port_number"])
        await vm.stop_capture(port_number)
        response.set_status(204)

    @Route.get(
        r"/projects/{project_id}/vpcs/nodes/{node_id}/adapters/{adapter_number:\d+}/ports/{port_number:\d+}/pcap",
        description="Stream the pcap capture file",
        parameters={
            "project_id": "Project UUID",
            "node_id": "Node UUID",
            "adapter_number": "Adapter to steam a packet capture",
            "port_number": "Port on the adapter"
        },
        status_codes={
            200: "File returned",
            403: "Permission denied",
            404: "The file doesn't exist"
        })
    async def stream_pcap_file(request, response):

        vpcs_manager = VPCS.instance()
        vm = vpcs_manager.get_node(request.match_info["node_id"], project_id=request.match_info["project_id"])
        port_number = int(request.match_info["port_number"])
        nio = vm.get_nio(port_number)
        await vpcs_manager.stream_pcap_file(nio, vm.project.id, request, response)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import asyncio
import sys

from tests.utils import asyncio_patch, AsyncioMagicMock
from unittest.mock import patch, MagicMock, ANY, PropertyMock

from gns3server.compute.traceng.traceng_vm import TraceNGVM
from gns3server.compute.traceng.traceng_error import TraceNGError
from gns3server.compute.traceng import TraceNG
from gns3server.compute.notification_manager import NotificationManager



@pytest.fixture
def manager(port_manager):
    m = TraceNG.instance()
    m.port_manager = port_manager
    return m


@pytest.fixture(scope="function")
def vm(project, manager, ubridge_path):
    vm = TraceNGVM("test", "00010203-0405-0607-0809-0a0b0c0d0e0f", project, manager)
    vm._start_ubridge = AsyncioMagicMock()
    vm._ubridge_hypervisor = MagicMock()
    vm._ubridge_hypervisor.is_running.return_value = True
    return vm


def test_vm(project, manager):
    vm = TraceNGVM("test", "00010203-0405-0607-0809-0a0b0c0d0e0f", project, manager)
    assert vm.name == "test"
    assert vm.id == "00010203-0405-0607-0809-0a0b0c0d0e0f"


def test_vm_invalid_traceng_path(vm, manager, loop):
    with patch("gns3server.compute.traceng.traceng_vm.TraceNGVM._traceng_path", return_value="/tmp/fake/path/traceng"):
        with pytest.raises(TraceNGError):
            nio = manager.create_nio({"type": "nio_udp", "lport": 4242, "rport": 4243, "rhost": "127.0.0.1"})
            loop.run_until_complete(asyncio.ensure_future(vm.port_add_nio_binding(0, nio)))
            loop.run_until_complete(asyncio.ensure_future(vm.start()))
            assert vm.name == "test"
            assert vm.id == "00010203-0405-0607-0809-0a0b0c0d0e0e"


def test_start(loop, vm, async_run):
    process = MagicMock()
    process.returncode = None

    with NotificationManager.instance().queue() as queue:
        async_run(queue.get(1))  # Ping

        vm.ip_address = "192.168.1.1"
        with patch("sys.platform", return_value="win"):
            with asyncio_patch("gns3server.compute.traceng.traceng_vm.TraceNGVM._check_requirements", return_value=True):
                with asyncio_patch("asyncio.create_subprocess_exec", return_value=process) as mock_exec:
                    loop.run_until_complete(asyncio.ensure_future(vm.start("192.168.1.2")))
                    assert mock_exec.call_args[0] == (vm._traceng_path(),
                                                      '-u',
                                                      '-c',
                                                      ANY,
                                                      '-v',
                                                      ANY,
                                                      '-b',
                                                      '127.0.0.1',
                                                      '-s',
                                                      'ICMP',
                                                      '-f',
                                                      '192.168.1.1',
                                                      '192.168.1.2')
                    assert vm.is_running()
                    assert vm.command_line == ' '.join(mock_exec.call_args[0])
        (action, event, kwargs) = async_run(queue.get(1))
        assert action == "node.updated"
        assert event == vm


def test_stop(loop, vm, async_run):
    process = MagicMock()

    # Wait process kill success
    future = asyncio.Future()
    future.set_result(True)
    process.wait.return_value = future
    process.returncode = None

    vm.ip_address = "192.168.1.1"
    with NotificationManager.instance().queue() as queue:
        with patch("sys.platform", return_value="win"):
            with asyncio_patch("gns3server.compute.traceng.traceng_vm.TraceNGVM._check_requirements", return_value=True):
                with asyncio_patch("asyncio.create_subprocess_exec", return_value=process):
                    nio = TraceNG.instance().create_nio({"type": "nio_udp", "lport": 4242, "rport": 4243, "rhost": "127.0.0.1", "filters": {}})
                    async_run(vm.port_add_nio_binding(0, nio))

                    vm._ubridge_send = AsyncioMagicMock()
                    async_run(vm.start("192.168.1.2"))
                    assert vm.is_running()

                    with asyncio_patch("gns3server.utils.asyncio.wait_for_process_termination"):
                        loop.run_until_complete(asyncio.ensure_future(vm.stop()))
                    assert vm.is_running() is False

                    process.terminate.assert_called_with()

                    async_run(queue.get(1))  #  Ping
                    async_run(queue.get(1))  #  Started

                    (action, event, kwargs) = async_run(queue.get(1))
                    assert action == "node.updated"
                    assert event == vm


def test_reload(loop, vm, async_run):
    process = MagicMock()

    # Wait process kill success
    future = asyncio.Future()
    future.set_result(True)
    process.wait.return_value = future
    process.returncode = None

    vm.ip_address = "192.168.1.1"
    with patch("sys.platform", return_value="win"):
        with asyncio_patch("gns3server.compute.traceng.traceng_vm.TraceNGVM._check_requirements", return_value=True):
            with asyncio_patch("asyncio.create_subprocess_exec", return_value=process):
                nio = TraceNG.instance().create_nio({"type": "nio_udp", "lport": 4242, "rport": 4243, "rhost": "127.0.0.1", "filters": {}})
                async_run(vm.port_add_nio_binding(0, nio))

                vm._ubridge_send = AsyncioMagicMock()
                async_run(vm.start("192.168.1.2"))
                assert vm.is_running()

                with asyncio_patch("gns3server.utils.asyncio.wait_for_process_termination"):
                    async_run(vm.reload())
                assert vm.is_running() is True

                #if sys.platform.startswith("win"):
                #    process.send_signal.assert_called_with(1)
                #else:
                process.terminate.assert_called_with()


def test_add_nio_binding_udp(vm, async_run):
    nio = TraceNG.instance().create_nio({"type": "nio_udp", "lport": 4242, "rport": 4243, "rhost": "127.0.0.1", "filters": {}})
    async_run(vm.port_add_nio_binding(0, nio))
    assert nio.lport == 4242


def test_port_remove_nio_binding(loop, vm):
    nio = TraceNG.instance().create_nio({"type": "nio_udp", "lport": 4242, "rport": 4243, "rhost": "127.0.0.1"})
    loop.run_until_complete(asyncio.ensure_future(vm.port_add_nio_binding(0, nio)))
    loop.run_until_complete(asyncio.ensure_future(vm.port_remove_nio_binding(0)))
    assert vm._ethernet_adapter.ports[0] is None


def test_close(vm, port_manager, loop):
    vm.ip_address = "192.168.1.1"
    with patch("sys.platform", return_value="win"):
        with asyncio_patch("gns3server.compute.traceng.traceng_vm.TraceNGVM._check_requirements", return_value=True):
            with asyncio_patch("asyncio.create_subprocess_exec", return_value=MagicMock()):
                loop.run_until_complete(asyncio.ensure_future(vm.start("192.168.1.2")))
                loop.run_until_complete(asyncio.ensure_future(vm.close()))
                assert vm.is_running() is False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Interface for Dynamips virtual Cisco 3600 instances module ("c3600")
http://github.com/GNS3/dynamips/blob/master/README.hypervisor#L366
"""

import asyncio
from .router import Router
from ..adapters.leopard_2fe import Leopard_2FE

import logging
log = logging.getLogger(__name__)


class C3600(Router):

    """
    Dynamips c3600 router.

    :param name: The name of this router
    :param node_id: Node identifier
    :param project: Project instance
    :param manager: Parent VM Manager
    :param dynamips_id: ID to use with Dynamips
    :param console: console port
    :param aux: auxiliary console port
    :param chassis: chassis for this router:
    3620, 3640 or 3660 (default = 3640).
    """

    def __init__(self, name, node_id, project, manager, dynamips_id, console=None, console_type="telnet", aux=None, chassis="3640"):

        super().__init__(name, node_id, project, manager, dynamips_id, console, console_type, aux, platform="c3600")

        # Set default values for this platform (must be the same as Dynamips)
        self._ram = 128
        self._nvram = 128
        self._disk0 = 0
        self._disk1 = 0
        self._iomem = 5  # percentage
        self._chassis = chassis
        self._clock_divisor = 4

    def __json__(self):

        c3600_router_info = {"iomem": self._iomem,
                             "chassis": self._chassis}

        router_info = Router.__json__(self)
        router_info.update(c3600_router_info)
        return router_info

    async def create(self):

        await Router.create(self)
        if self._chassis != "3640":
            await self.set_chassis(self._chassis)
        self._setup_chassis()

    def _setup_chassis(self):
        """
        Sets up the router with the corresponding chassis
        (create slots and insert default adapters).
        """

        if self._chassis == "3620":
            self._create_slots(2)
        elif self._chassis == "3640":
            self._create_slots(4)
        elif self._chassis == "3660":
            self._create_slots(7)
            self._slots[0] = Leopard_2FE()

    @property
    def chassis(self):
        """
        Returns the chassis.

        :returns: chassis string
        """

        return self._chassis

    async def set_chassis(self, chassis):
        """
        Sets the chassis.

        :param: chassis string: 3620, 3640 or 3660
        """

        await self._hypervisor.send('c3600 set_chassis "{name}" {chassis}'.format(name=self._name, chassis=chassis))

        log.info('Router "{name}" [{id}]: chassis set to {chassis}'.format(name=self._name,
                                                                           id=self._id,
                                                                           chassis=chassis))

        self._chassis = chassis
        self._setup_chassis()

    @property
    def iomem(self):
        """
        Returns I/O memory size for this router.

        :returns: I/O memory size (integer)
        """

        return self._iomem

    async def set_iomem(self, iomem):
        """
        Set I/O memory size for this router.

        :param iomem: I/O memory size
        """

        await self._hypervisor.send('c3600 set_iomem "{name}" {size}'.format(name=self._name, size=iomem))

        log.info('Router "{name}" [{id}]: I/O memory updated from {old_iomem}% to {new_iomem}%'.format(name=self._name,
                                                                                                       id=self._id,
                                                                                                       old_iomem=self._iomem,
                                                                                                       new_iomem=iomem))
        self._iomem = iomem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from unittest.mock import MagicMock
from aiohttp.web import HTTPNotFound

from gns3server.web.response import Response


@pytest.fixture()
def response():
    request = MagicMock()
    return Response(request=request)


def test_response_file(async_run, tmpdir, response):
    filename = str(tmpdir / 'hello')
    with open(filename, 'w+') as f:
        f.write('world')

    async_run(response.stream_file(filename))
    assert response.status == 200


def test_response_file_not_found(async_run, tmpdir, response):
    filename = str(tmpdir / 'hello-not-found')

    pytest.raises(HTTPNotFound, lambda: async_run(response.stream_file(filename)))



#!/usr/bin/env python
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
from .label import LABEL_OBJECT_SCHEMA
from .custom_adapters import CUSTOM_ADAPTERS_ARRAY_SCHEMA

NODE_TYPE_SCHEMA = {
    "description": "Type of node",
    "enum": [
        "cloud",
        "nat",
        "ethernet_hub",
        "ethernet_switch",
        "frame_relay_switch",
        "atm_switch",
        "docker",
        "dynamips",
        "vpcs",
        "traceng",
        "virtualbox",
        "vmware",
        "iou",
        "qemu"
    ]
}

NODE_LIST_IMAGES_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "List of binary images",
    "type": "array",
    "items": [
        {
            "type": "object",
            "properties": {
                "filename": {
                    "description": "Image filename",
                    "type": "string",
                    "minLength": 1
                },
                "path": {
                    "description": "Image path",
                    "type": "string",
                    "minLength": 1
                },
                "md5sum": {
                    "description": "md5sum of the image if available",
                    "type": ["string", "null"],
                    "minLength": 1
                },
                "filesize": {
                    "description": "size of the image if available",
                    "type": ["integer", "null"],
                    "minimum": 0
                }
            },
            "required": ["filename", "path"],
            "additionalProperties": False
        }
    ],
    "additionalProperties": False,
}


NODE_CAPTURE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Request validation to start a packet capture on a port",
    "type": "object",
    "properties": {
        "capture_file_name": {
            "description": "Capture file name",
            "type": "string",
            "minLength": 1,
        },
        "data_link_type": {
            "description": "PCAP data link type (http://www.tcpdump.org/linktypes.html)",
            "enum": ["DLT_ATM_RFC1483", "DLT_EN10MB", "DLT_FRELAY", "DLT_C_HDLC", "DLT_PPP_SERIAL"]
        }
    },
    "additionalProperties": False,
    "required": ["capture_file_name"]
}


NODE_OBJECT_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "A node object",
    "type": "object",
    "properties": {
        "compute_id": {
            "description": "Compute identifier",
            "type": "string"
        },
        "project_id": {
            "description": "Project UUID",
            "type": "string",
            "minLength": 36,
            "maxLength": 36,
            "pattern": "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
        },
        "node_id": {
            "description": "Node UUID",
            "type": "string",
            "minLength": 36,
            "maxLength": 36,
            "pattern": "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
        },
        "template_id": {
            "description": "Template UUID from which the node has been created. Read only",
            "type": ["null", "string"],
            "minLength": 36,
            "maxLength": 36,
            "pattern": "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
        },
        "node_type": NODE_TYPE_SCHEMA,
        "node_directory": {
            "description": "Working directory of the node. Read only",
            "type": ["null", "string"]
        },
        "command_line": {
            "description": "Command line use to start the node",
            "type": ["null", "string"]
        },
        "name": {
            "description": "Node name",
            "type": "string",
            "minLength": 1,
        },
        "console": {
            "description": "Console TCP port",
            "minimum": 1,
            "maximum": 65535,
            "type": ["integer", "null"]
        },
        "console_host": {
            "description": "Console host. Warning if the host is 0.0.0.0 or :: (listen on all interfaces) you need to use the same address you use to connect to the controller.",
            "type": "string",
            "minLength": 1,
        },
        "console_type": {
            "description": "Console type",
            "enum": ["vnc", "telnet", "http", "https", "spice", "spice+agent", "none", None]
        },
        "console_auto_start": {
            "description": "Automatically start the console when the node has started",
            "type": "boolean"
        },
        "properties": {
            "description": "Properties specific to an emulator",
            "type": "object"
        },
        "status": {
            "description": "Status of the node",
            "enum": ["stopped", "started", "suspended"]
        },
        "label": LABEL_OBJECT_SCHEMA,
        "symbol": {
            "description": "Symbol of the node",
            "type": ["string", "null"],
            "minLength": 1
        },
        "width": {
            "description": "Width of the node (Read only)",
            "type": "integer"
        },
        "height": {
            "description": "Height of the node (Read only)",
            "type": "integer"
        },
        "x": {
            "description": "X position of the node",
            "type": "integer"
        },
        "y": {
            "description": "Y position of the node",
            "type": "integer"
        },
        "z": {
            "description": "Z position of the node",
            "type": "integer"
        },
        "locked": {
            "description": "Whether the element locked or not",
            "type": "boolean"
        },
        "port_name_format": {
            "description": "Formating for port name {0} will be replace by port number",
            "type": "string"
        },
        "port_segment_size": {
            "description": "Size of the port segment",
            "type": "integer",
            "minimum": 0
        },
        "first_port_name": {
            "description": "Name of the first port",
            "type": ["string", "null"],
        },
        "custom_adapters": CUSTOM_ADAPTERS_ARRAY_SCHEMA,
        "ports": {
            "description": "List of node ports READ only",
            "type": "array",
            "items": {
                "type": "object",
                "description": "A node port",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "Port name",
                    },
                    "short_name": {
                        "type": "string",
                        "description": "Short version of port name",
                    },
                    "adapter_number": {
                        "type": "integer",
                        "description": "Adapter slot"
                    },
                    "adapter_type": {
                        "description": "Adapter type",
                        "type": ["string", "null"],
                        "minLength": 1,
                    },
                    "port_number": {
                        "type": "integer",
                        "description": "Port slot"
                    },
                    "link_type": {
                        "description": "Type of link",
                        "enum": ["ethernet", "serial"]
                    },
                    "data_link_types": {
                        "type": "object",
                        "description": "Available PCAP types for capture",
                        "properties": {}
                    },
                    "mac_address": {
                        "description": "MAC address (if available)",
                        "type": ["string", "null"],
                        "minLength": 1,
                        "pattern": "^([0-9a-fA-F]{2}[:]){5}([0-9a-fA-F]{2})$"
                    },
                },
                "additionalProperties": False
            }
        }
    },
    "additionalProperties": False,
    "required": ["name", "node_type", "compute_id"]
}

NODE_CREATE_SCHEMA = NODE_OBJECT_SCHEMA
NODE_UPDATE_SCHEMA = copy.deepcopy(NODE_OBJECT_SCHEMA)
del NODE_UPDATE_SCHEMA["required"]


NODE_DUPLICATE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Duplicate a node",
    "type": "object",
    "properties": {
        "x": {
            "description": "X position of the node",
            "type": "integer"
        },
        "y": {
            "description": "Y position of the node",
            "type": "integer"
        },
        "z": {
            "description": "Z position of the node",
            "type": "integer"
        }
    },
    "additionalProperties": False,
    "required": ["x", "y"]
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pytest
import uuid


from gns3server.compute.vpcs import VPCS
from gns3server.compute.vpcs.vpcs_error import VPCSError
from gns3server.compute.project_manager import ProjectManager


def test_get_mac_id(loop, project, port_manager):
    # Cleanup the VPCS object
    VPCS._instance = None
    vpcs = VPCS.instance()
    vpcs.port_manager = port_manager
    vm1_id = str(uuid.uuid4())
    vm2_id = str(uuid.uuid4())
    vm3_id = str(uuid.uuid4())
    loop.run_until_complete(vpcs.create_node("PC 1", project.id, vm1_id))
    loop.run_until_complete(vpcs.create_node("PC 2", project.id, vm2_id))
    assert vpcs.get_mac_id(vm1_id) == 0
    assert vpcs.get_mac_id(vm1_id) == 0
    assert vpcs.get_mac_id(vm2_id) == 1
    loop.run_until_complete(vpcs.delete_node(vm1_id))
    loop.run_until_complete(vpcs.create_node("PC 3", project.id, vm3_id))
    assert vpcs.get_mac_id(vm3_id) == 0


def test_get_mac_id_multiple_project(loop, port_manager):
    # Cleanup the VPCS object
    VPCS._instance = None
    vpcs = VPCS.instance()
    vpcs.port_manager = port_manager
    vm1_id = str(uuid.uuid4())
    vm2_id = str(uuid.uuid4())
    vm3_id = str(uuid.uuid4())
    project1 = ProjectManager.instance().create_project(project_id=str(uuid.uuid4()))
    project2 = ProjectManager.instance().create_project(project_id=str(uuid.uuid4()))
    loop.run_until_complete(vpcs.create_node("PC 1", project1.id, vm1_id))
    loop.run_until_complete(vpcs.create_node("PC 2", project1.id, vm2_id))
    loop.run_until_complete(vpcs.create_node("PC 2", project2.id, vm3_id))
    assert vpcs.get_mac_id(vm1_id) == 0
    assert vpcs.get_mac_id(vm2_id) == 1
    assert vpcs.get_mac_id(vm3_id) == 0


def test_get_mac_id_no_id_available(loop, project, port_manager):
    # Cleanup the VPCS object
    VPCS._instance = None
    vpcs = VPCS.instance()
    vpcs.port_manager = port_manager
    with pytest.raises(VPCSError):
        for i in range(0, 256):
            node_id = str(uuid.uuid4())
            loop.run_until_complete(vpcs.create_node("PC {}".format(i), project.id, node_id))
            assert vpcs.get_mac_id(node_id) == i



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import functools
import asyncio
import sys
import os
import threading

from asyncio.futures import CancelledError


async def wait_run_in_executor(func, *args, **kwargs):
    """
    Run blocking code in a different thread and wait
    for the result.

    :param func: Run this function in a different thread
    :param args: Parameters of the function
    :param kwargs: Keyword parameters of the function
    :returns: Return the result of the function
    """

    loop = asyncio.get_event_loop()
    future = loop.run_in_executor(None, functools.partial(func, *args, **kwargs))
    await asyncio.wait([future])
    return future.result()


async def cancellable_wait_run_in_executor(func, *args, **kwargs):
    """
    Run blocking code in a different thread and wait
    for the result. Support cancellation.

    :param func: Run this function in a different thread
    :param args: Parameters of the function
    :param kwargs: Keyword parameters of the function
    :returns: Return the result of the function
    """
    stopped_event = threading.Event()
    kwargs['stopped_event'] = stopped_event
    try:
        await wait_run_in_executor(func, *args, **kwargs)
    except CancelledError:
        stopped_event.set()


async def subprocess_check_output(*args, cwd=None, env=None, stderr=False):
    """
    Run a command and capture output

    :param *args: List of command arguments
    :param cwd: Current working directory
    :param env: Command environment
    :param stderr: Read on stderr
    :returns: Command output
    """

    if stderr:
        proc = await asyncio.create_subprocess_exec(*args, stderr=asyncio.subprocess.PIPE, cwd=cwd, env=env)
        output = await proc.stderr.read()
    else:
        proc = await asyncio.create_subprocess_exec(*args, stdout=asyncio.subprocess.PIPE, cwd=cwd, env=env)
        output = await proc.stdout.read()
    if output is None:
        return ""
    # If we received garbage we ignore invalid characters
    # it should happens only when user try to use another binary
    # and the code of VPCS, dynamips... Will detect it's not the correct binary
    return output.decode("utf-8", errors="ignore")

async def wait_for_process_termination(process, timeout=10):
    """
    Wait for a process terminate, and raise asyncio.TimeoutError in case of
    timeout.

    In theory this can be implemented by just:
    await asyncio.wait_for(self._iou_process.wait(), timeout=100)

    But it's broken before Python 3.4:
    http://bugs.python.org/issue23140

    :param process: An asyncio subprocess
    :param timeout: Timeout in seconds
    """

    if sys.version_info >= (3, 5):
        try:
            await asyncio.wait_for(process.wait(), timeout=timeout)
        except ProcessLookupError:
            return
    else:
        while timeout > 0:
            if process.returncode is not None:
                return
            await asyncio.sleep(0.1)
            timeout -= 0.1
        raise asyncio.TimeoutError()


async def _check_process(process, termination_callback):
    if not hasattr(sys, "_called_from_test") or not sys._called_from_test:
        returncode = await process.wait()
        if asyncio.iscoroutinefunction(termination_callback):
            await termination_callback(returncode)
        else:
            termination_callback(returncode)


def monitor_process(process, termination_callback):
    """Call termination_callback when a process dies"""

    asyncio.ensure_future(_check_process(process, termination_callback))


async def wait_for_file_creation(path, timeout=10):

    while timeout > 0:
        if os.path.exists(path):
            return
        await asyncio.sleep(0.5)
        timeout -= 0.5
    raise asyncio.TimeoutError()


async def wait_for_named_pipe_creation(pipe_path, timeout=60):

    import win32pipe
    import pywintypes

    while timeout > 0:
        try:
            win32pipe.WaitNamedPipe(pipe_path, 1)
        except pywintypes.error:
            await asyncio.sleep(0.5)
            timeout -= 0.5
        else:
            return
    raise asyncio.TimeoutError()


def locking(f):

    @functools.wraps(f)
    async def wrapper(oself, *args, **kwargs):
        lock_name = "__" + f.__name__ + "_lock"
        if not hasattr(oself, lock_name):
            setattr(oself, lock_name, asyncio.Lock())
        async with getattr(oself, lock_name):
            return await f(oself, *args, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import hashlib

from ..config import Config
from . import force_unix_path


import logging
log = logging.getLogger(__name__)


def list_images(type):
    """
    Scan directories for available image for a type

    :param type: emulator type (dynamips, qemu, iou)
    """
    files = set()
    images = []

    server_config = Config.instance().get_section_config("Server")
    general_images_directory = os.path.expanduser(server_config.get("images_path", "~/GNS3/images"))

    # Subfolder of the general_images_directory specific to this VM type
    default_directory = default_images_directory(type)

    for directory in images_directories(type):

        # We limit recursion to path outside the default images directory
        # the reason is in the default directory manage file organization and
        # it should be flatten to keep things simple
        recurse = True
        if os.path.commonprefix([directory, general_images_directory]) == general_images_directory:
            recurse = False

        directory = os.path.normpath(directory)
        for root, _, filenames in _os_walk(directory, recurse=recurse):
            for filename in filenames:
                path = os.path.join(root, filename)
                if filename not in files:
                    if filename.endswith(".md5sum") or filename.startswith("."):
                        continue
                    elif ((filename.endswith(".image") or filename.endswith(".bin")) and type == "dynamips") \
                            or ((filename.endswith(".bin") or filename.startswith("i86bi")) and type == "iou") \
                            or (not filename.endswith(".bin") and not filename.endswith(".image") and type == "qemu"):
                        files.add(filename)

                        # It the image is located in the standard directory the path is relative
                        if os.path.commonprefix([root, default_directory]) != default_directory:
                            path = os.path.join(root, filename)
                        else:
                            path = os.path.relpath(os.path.join(root, filename), default_directory)

                        try:
                            if type in ["dynamips", "iou"]:
                                with open(os.path.join(root, filename), "rb") as f:
                                    # read the first 7 bytes of the file.
                                    elf_header_start = f.read(7)
                                # valid IOS images must start with the ELF magic number, be 32-bit, big endian and have an ELF version of 1
                                if not elf_header_start == b'\x7fELF\x01\x02\x01' and not elf_header_start == b'\x7fELF\x01\x01\x01':
                                    continue

                            images.append({
                                "filename": filename,
                                "path": force_unix_path(path),
                                "md5sum": md5sum(os.path.join(root, filename)),
                                "filesize": os.stat(os.path.join(root, filename)).st_size})
                        except OSError as e:
                            log.warning("Can't add image {}: {}".format(path, str(e)))
    return images


def _os_walk(directory, recurse=True, **kwargs):
    """
    Work like os.walk but if recurse is False just list current directory
    """
    if recurse:
        for root, dirs, files in os.walk(directory, **kwargs):
            yield root, dirs, files
    else:
        files = []
        for filename in os.listdir(directory):
            if os.path.isfile(os.path.join(directory, filename)):
                files.append(filename)
        yield directory, [], files


def default_images_directory(type):
    """
    :returns: Return the default directory for a node type
    """
    server_config = Config.instance().get_section_config("Server")
    img_dir = os.path.expanduser(server_config.get("images_path", "~/GNS3/images"))
    if type == "qemu":
        return os.path.join(img_dir, "QEMU")
    elif type == "iou":
        return os.path.join(img_dir, "IOU")
    elif type == "dynamips":
        return os.path.join(img_dir, "IOS")
    else:
        raise NotImplementedError("%s node type is not supported", type)


def images_directories(type):
    """
    Return all directories where we will look for images
    by priority

    :param type: Type of emulator
    """
    server_config = Config.instance().get_section_config("Server")

    paths = []
    img_dir = os.path.expanduser(server_config.get("images_path", "~/GNS3/images"))
    type_img_directory = default_images_directory(type)
    try:
        os.makedirs(type_img_directory, exist_ok=True)
        paths.append(type_img_directory)
    except (OSError, PermissionError):
        pass
    for directory in server_config.get("additional_images_path", "").split(";"):
        paths.append(directory)
    # Compatibility with old topologies we look in parent directory
    paths.append(img_dir)
    # Return only the existing paths
    return [force_unix_path(p) for p in paths if os.path.exists(p)]


def md5sum(path, stopped_event=None):
    """
    Return the md5sum of an image and cache it on disk

    :param path: Path to the image
    :param stopped_event: In case you execute this function on thread and would like to have possibility
                          to cancel operation pass the `threading.Event`
    :returns: Digest of the image
    """

    if path is None or len(path) == 0 or not os.path.exists(path):
        return None

    try:
        with open(path + '.md5sum') as f:
            md5 = f.read().strip()
            if len(md5) == 32:
                return md5
    # Unicode error is when user rename an image to .md5sum ....
    except (OSError, UnicodeDecodeError):
        pass

    try:
        m = hashlib.md5()
        with open(path, 'rb') as f:
            while True:
                if stopped_event is not None and stopped_event.is_set():
                    log.error("MD5 sum calculation of `{}` has stopped due to cancellation".format(path))
                    return
                buf = f.read(128)
                if not buf:
                    break
                m.update(buf)
        digest = m.hexdigest()
    except OSError as e:
        log.error("Can't create digest of %s: %s", path, str(e))
        return None

    try:
        with open('{}.md5sum'.format(path), 'w+') as f:
            f.write(digest)
    except OSError as e:
        log.error("Can't write digest of %s: %s", path, str(e))

    return digest


def remove_checksum(path):
    """
    Remove the checksum of an image from cache if exists
    """

    path = '{}.md5sum'.format(path)
    if os.path.exists(path):
        os.remove(path)



#!/usr/bin/env python
#
# Copyright (C) 2016 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import uuid


class Appliance:

    def __init__(self, appliance_id, data, builtin=True):
        if appliance_id is None:
            self._id = str(uuid.uuid4())
        elif isinstance(appliance_id, uuid.UUID):
            self._id = str(appliance_id)
        else:
            self._id = appliance_id
        self._data = data.copy()
        self._builtin = builtin
        if "appliance_id" in self._data:
            del self._data["appliance_id"]

    @property
    def id(self):
        return self._id

    @property
    def status(self):
        return self._data["status"]

    @property
    def symbol(self):
        return self._data.get("symbol")

    @symbol.setter
    def symbol(self, new_symbol):
        self._data["symbol"] = new_symbol

    def __json__(self):
        """
        Appliance data (a hash)
        """
        data = copy.deepcopy(self._data)
        data["builtin"] = self._builtin
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gns3server.web.route import Route
from gns3server.compute.port_manager import PortManager
from gns3server.compute.project_manager import ProjectManager
from gns3server.utils.interfaces import interfaces


class NetworkHandler:

    @Route.post(
        r"/projects/{project_id}/ports/udp",
        parameters={
            "project_id": "Project UUID",
        },
        status_codes={
            201: "UDP port allocated",
            404: "The project doesn't exist"
        },
        description="Allocate an UDP port on the server")
    def allocate_udp_port(request, response):

        pm = ProjectManager.instance()
        project = pm.get_project(request.match_info["project_id"])
        m = PortManager.instance()
        udp_port = m.get_free_udp_port(project)
        response.set_status(201)
        response.json({"udp_port": udp_port})

    @Route.get(
        r"/network/interfaces",
        description="List all the network interfaces available on the server")
    def network_interfaces(request, response):

        network_interfaces = interfaces()
        response.json(network_interfaces)

    @Route.get(
        r"/network/ports",
        description="List all the ports used by the server")
    def network_ports(request, response):

        m = PortManager.instance()
        response.json(m)



#!/usr/bin/env python
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .label import LABEL_OBJECT_SCHEMA
from .filter import FILTER_OBJECT_SCHEMA

LINK_OBJECT_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "A link object",
    "type": "object",
    "properties": {
        "link_id": {
            "description": "Link UUID",
            "type": "string",
            "minLength": 36,
            "maxLength": 36,
            "pattern": "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
        },
        "project_id": {
            "description": "Project UUID",
            "type": "string",
            "minLength": 36,
            "maxLength": 36,
            "pattern": "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
        },
        "nodes": {
            "description": "List of the VMS",
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "node_id": {
                        "description": "Node UUID",
                        "type": "string",
                        "minLength": 36,
                        "maxLength": 36,
                        "pattern": "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
                    },
                    "adapter_number": {
                        "description": "Adapter number",
                        "type": "integer"
                    },
                    "port_number": {
                        "description": "Port number",
                        "type": "integer"
                    },
                    "label": LABEL_OBJECT_SCHEMA
                },
                "required": ["node_id", "adapter_number", "port_number"],
                "additionalProperties": False
            }
        },
        "suspend": {
            "type": "boolean",
            "description": "Suspend the link"
        },
        "filters": FILTER_OBJECT_SCHEMA,
        "capturing": {
            "description": "Read only property. True if a capture running on the link",
            "type": "boolean"
        },
        "capture_file_name": {
            "description": "Read only property. The name of the capture file if a capture is running",
            "type": ["string", "null"]
        },
        "capture_file_path": {
            "description": "Read only property. The full path of the capture file if a capture is running",
            "type": ["string", "null"]
        },
        "capture_compute_id": {
            "description": "Read only property. The compute identifier where a capture is running",
            "type": ["string", "null"]
        },
        "link_type": {
            "description": "Type of link",
            "enum": ["ethernet", "serial"]
        }
    },
    "additionalProperties": False
}


LINK_CAPTURE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Request validation to start a packet capture on a link",
    "type": "object",
    "properties": {
        "data_link_type": {
            "description": "PCAP data link type (http://www.tcpdump.org/linktypes.html)",
            "enum": ["DLT_ATM_RFC1483", "DLT_EN10MB", "DLT_FRELAY", "DLT_C_HDLC", "DLT_PPP_SERIAL"]
        },
        "capture_file_name": {
            "description": "Read only property. The name of the capture file if capture is running",
            "type": "string"
        }
    },
    "additionalProperties": False
}



#!/usr/bin/env python
#
# Copyright (C) 2016 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import aiohttp
import asyncio

from gns3server.web.route import Route
from gns3server.controller import Controller


import logging
log = logging.getLogger(__name__)


class SymbolHandler:
    """
    API entry points for symbols management.
    """

    @Route.get(
        r"/symbols",
        description="List of symbols",
        status_codes={
            200: "Symbols list returned"
        })
    def list(request, response):

        controller = Controller.instance()
        response.json(controller.symbols.list())

    @Route.get(
        r"/symbols/{symbol_id:.+}/raw",
        description="Get the symbol file",
        status_codes={
            200: "Symbol returned"
        })
    async def raw(request, response):

        controller = Controller.instance()
        try:
            await response.stream_file(controller.symbols.get_path(request.match_info["symbol_id"]))
        except (KeyError, OSError) as e:
            log.warning("Could not get symbol file: {}".format(e))
            response.set_status(404)

    @Route.post(
        r"/symbols/{symbol_id:.+}/raw",
        description="Write the symbol file",
        status_codes={
            200: "Symbol written"
        },
        raw=True)
    async def upload(request, response):
        controller = Controller.instance()
        path = os.path.join(controller.symbols.symbols_path(), os.path.basename(request.match_info["symbol_id"]))
        try:
            with open(path, "wb") as f:
                while True:
                    try:
                        chunk = await request.content.read(1024)
                    except asyncio.TimeoutError:
                        raise aiohttp.web.HTTPRequestTimeout(text="Timeout when writing to symbol '{}'".format(path))
                    if not chunk:
                        break
                    f.write(chunk)
        except (UnicodeEncodeError, OSError) as e:
            raise aiohttp.web.HTTPConflict(text="Could not write symbol file '{}': {}".format(path, e))

        # Reset the symbol list
        controller.symbols.list()
        response.set_status(204)

    @Route.get(
        r"/default_symbols",
        description="List of default symbols",
        status_codes={
            200: "Default symbols list returned"
        })
    def list_default_symbols(request, response):

        controller = Controller.instance()
        response.json(controller.symbols.default_symbols())

    # @Route.post(
    #     r"/symbol_theme",
    #     description="Create a new symbol theme",
    #     status_codes={
    #         201: "Appliance created",
    #         400: "Invalid request"
    #     },
    #     input=APPLIANCE_CREATE_SCHEMA,
    #     output=APPLIANCE_OBJECT_SCHEMA)
    # def create(request, response):
    #
    #     controller = Controller.instance()
    #     appliance = controller.add_appliance(request.json)
    #     response.set_status(201)
    #     response.json(appliance)


# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
TraceNG server module.
"""

import asyncio

from ..base_manager import BaseManager
from .traceng_error import TraceNGError
from .traceng_vm import TraceNGVM


class TraceNG(BaseManager):

    _NODE_CLASS = TraceNGVM

    def __init__(self):

        super().__init__()

    async def create_node(self, *args, **kwargs):
        """
        Creates a new TraceNG VM.

        :returns: TraceNGVM instance
        """

        return (await super().create_node(*args, **kwargs))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..qemu_error import QemuError

import logging
log = logging.getLogger(__name__)


def get_next_guest_cid(nodes):
    """
    Calculates free guest_id from given nodes

    :param nodes:
    :raises QemuError when exceeds number
    :return: integer first free cid
    """

    used = set([n.guest_cid for n in nodes])
    pool = set(range(3, 65535))
    try:
        return (pool - used).pop()
    except KeyError:
        raise QemuError("Cannot create a new Qemu VM (limit of 65535 guest ID on one host reached)")



#!/usr/bin/env python
#
# Copyright (C) 2016 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import uuid
import pytest
from unittest.mock import MagicMock
from tests.utils import asyncio_patch

from gns3server.controller import Controller
from gns3server.controller.template import Template


def test_template_list(http_controller, controller):

    id = str(uuid.uuid4())
    controller.template_manager.load_templates()
    controller.template_manager._templates[id] = Template(id, {
        "template_type": "qemu",
        "category": 0,
        "name": "test",
        "symbol": "guest.svg",
        "default_name_format": "{name}-{0}",
        "compute_id": "local"
    })
    response = http_controller.get("/templates", example=True)
    assert response.status == 200
    assert response.route == "/templates"
    assert len(response.json) > 0


def test_template_create_without_id(http_controller, controller):

    params = {"base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params, example=True)
    assert response.status == 201
    assert response.route == "/templates"
    assert response.json["template_id"] is not None
    assert len(controller.template_manager._templates) == 1


def test_template_create_with_id(http_controller, controller):

    params = {"template_id": str(uuid.uuid4()),
              "base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params, example=True)
    assert response.status == 201
    assert response.route == "/templates"
    assert response.json["template_id"] is not None
    assert len(controller.template_manager._templates) == 1


def test_template_create_wrong_type(http_controller, controller):

    params = {"template_id": str(uuid.uuid4()),
              "base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "invalid_template_type"}

    response = http_controller.post("/templates", params)
    assert response.status == 400
    assert len(controller.template_manager._templates) == 0


def test_template_get(http_controller, controller):

    template_id = str(uuid.uuid4())
    params = {"template_id": template_id,
              "base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params)
    assert response.status == 201

    response = http_controller.get("/templates/{}".format(template_id), example=True)
    assert response.status == 200
    assert response.json["template_id"] == template_id


def test_template_update(http_controller, controller):

    template_id = str(uuid.uuid4())
    params = {"template_id": template_id,
              "base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params)
    assert response.status == 201

    response = http_controller.get("/templates/{}".format(template_id))
    assert response.status == 200
    assert response.json["template_id"] == template_id

    params["name"] = "VPCS_TEST_RENAMED"
    response = http_controller.put("/templates/{}".format(template_id), params, example=True)

    assert response.status == 200
    assert response.json["name"] == "VPCS_TEST_RENAMED"


def test_template_delete(http_controller, controller):

    template_id = str(uuid.uuid4())
    params = {"template_id": template_id,
              "base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params)
    assert response.status == 201

    response = http_controller.get("/templates")
    assert len(response.json) == 1
    assert len(controller.template_manager._templates) == 1

    response = http_controller.delete("/templates/{}".format(template_id), example=True)
    assert response.status == 204

    response = http_controller.get("/templates")
    assert len(response.json) == 0
    assert len(controller.template_manager._templates) == 0


def test_template_duplicate(http_controller, controller):

    template_id = str(uuid.uuid4())
    params = {"template_id": template_id,
              "base_script_file": "vpcs_base_config.txt",
              "category": "guest",
              "console_auto_start": False,
              "console_type": "telnet",
              "default_name_format": "PC{0}",
              "name": "VPCS_TEST",
              "compute_id": "local",
              "symbol": ":/symbols/vpcs_guest.svg",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params)
    assert response.status == 201

    response = http_controller.post("/templates/{}/duplicate".format(template_id), example=True)
    assert response.status == 201
    assert response.json["template_id"] != template_id
    params.pop("template_id")
    for param, value in params.items():
        assert response.json[param] == value

    response = http_controller.get("/templates")
    assert len(response.json) == 2
    assert len(controller.template_manager._templates) == 2


def test_c7200_dynamips_template_create(http_controller):

    params = {"name": "Cisco c7200 template",
              "platform": "c7200",
              "compute_id": "local",
              "image": "c7200-adventerprisek9-mz.124-24.T5.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c7200-adventerprisek9-mz.124-24.T5.image",
                         "mac_addr": "",
                         "midplane": "vxr",
                         "mmap": True,
                         "name": "Cisco c7200 template",
                         "npe": "npe-400",
                         "nvram": 512,
                         "platform": "c7200",
                         "private_config": "",
                         "ram": 512,
                         "sparsemem": True,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c3745_dynamips_template_create(http_controller):

    params = {"name": "Cisco c3745 template",
              "platform": "c3745",
              "compute_id": "local",
              "image": "c3745-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c3745-adventerprisek9-mz.124-25d.image",
                         "mac_addr": "",
                         "mmap": True,
                         "name": "Cisco c3745 template",
                         "iomem": 5,
                         "nvram": 256,
                         "platform": "c3745",
                         "private_config": "",
                         "ram": 256,
                         "sparsemem": True,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c3725_dynamips_template_create(http_controller):

    params = {"name": "Cisco c3725 template",
              "platform": "c3725",
              "compute_id": "local",
              "image": "c3725-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c3725-adventerprisek9-mz.124-25d.image",
                         "mac_addr": "",
                         "mmap": True,
                         "name": "Cisco c3725 template",
                         "iomem": 5,
                         "nvram": 256,
                         "platform": "c3725",
                         "private_config": "",
                         "ram": 128,
                         "sparsemem": True,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c3600_dynamips_template_create(http_controller):

    params = {"name": "Cisco c3600 template",
              "platform": "c3600",
              "chassis": "3660",
              "compute_id": "local",
              "image": "c3660-a3jk9s-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c3660-a3jk9s-mz.124-25d.image",
                         "mac_addr": "",
                         "mmap": True,
                         "name": "Cisco c3600 template",
                         "iomem": 5,
                         "nvram": 128,
                         "platform": "c3600",
                         "chassis": "3660",
                         "private_config": "",
                         "ram": 192,
                         "sparsemem": True,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c3600_dynamips_template_create_wrong_chassis(http_controller):

    params = {"name": "Cisco c3600 template",
              "platform": "c3600",
              "chassis": "3650",
              "compute_id": "local",
              "image": "c3660-a3jk9s-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 400


def test_c2691_dynamips_template_create(http_controller):

    params = {"name": "Cisco c2691 template",
              "platform": "c2691",
              "compute_id": "local",
              "image": "c2691-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c2691-adventerprisek9-mz.124-25d.image",
                         "mac_addr": "",
                         "mmap": True,
                         "name": "Cisco c2691 template",
                         "iomem": 5,
                         "nvram": 256,
                         "platform": "c2691",
                         "private_config": "",
                         "ram": 192,
                         "sparsemem": True,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c2600_dynamips_template_create(http_controller):

    params = {"name": "Cisco c2600 template",
              "platform": "c2600",
              "chassis": "2651XM",
              "compute_id": "local",
              "image": "c2600-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c2600-adventerprisek9-mz.124-25d.image",
                         "mac_addr": "",
                         "mmap": True,
                         "name": "Cisco c2600 template",
                         "iomem": 15,
                         "nvram": 128,
                         "platform": "c2600",
                         "chassis": "2651XM",
                         "private_config": "",
                         "ram": 160,
                         "sparsemem": True,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c2600_dynamips_template_create_wrong_chassis(http_controller):

    params = {"name": "Cisco c2600 template",
              "platform": "c2600",
              "chassis": "2660XM",
              "compute_id": "local",
              "image": "c2600-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 400


def test_c1700_dynamips_template_create(http_controller):

    params = {"name": "Cisco c1700 template",
              "platform": "c1700",
              "chassis": "1760",
              "compute_id": "local",
              "image": "c1700-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "dynamips",
                         "auto_delete_disks": False,
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "R{0}",
                         "disk0": 0,
                         "disk1": 0,
                         "exec_area": 64,
                         "idlemax": 500,
                         "idlepc": "",
                         "idlesleep": 30,
                         "image": "c1700-adventerprisek9-mz.124-25d.image",
                         "mac_addr": "",
                         "mmap": True,
                         "name": "Cisco c1700 template",
                         "iomem": 15,
                         "nvram": 128,
                         "platform": "c1700",
                         "chassis": "1760",
                         "private_config": "",
                         "ram": 160,
                         "sparsemem": False,
                         "startup_config": "ios_base_startup-config.txt",
                         "symbol": ":/symbols/router.svg",
                         "system_id": "FTX0945W0MY"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_c1700_dynamips_template_create_wrong_chassis(http_controller):

    params = {"name": "Cisco c1700 template",
              "platform": "c1700",
              "chassis": "1770",
              "compute_id": "local",
              "image": "c1700-adventerprisek9-mz.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 400


def test_dynamips_template_create_wrong_platform(http_controller):

    params = {"name": "Cisco c3900 template",
              "platform": "c3900",
              "compute_id": "local",
              "image": "c3900-test.124-25d.image",
              "template_type": "dynamips"}

    response = http_controller.post("/templates", params)
    assert response.status == 400


def test_iou_template_create(http_controller):

    params = {"name": "IOU template",
              "compute_id": "local",
              "path": "/path/to/i86bi_linux-ipbase-ms-12.4.bin",
              "template_type": "iou"}

    response = http_controller.post("/templates", params, example=True)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "iou",
                         "builtin": False,
                         "category": "router",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "IOU{0}",
                         "ethernet_adapters": 2,
                         "name": "IOU template",
                         "nvram": 128,
                         "path": "/path/to/i86bi_linux-ipbase-ms-12.4.bin",
                         "private_config": "",
                         "ram": 256,
                         "serial_adapters": 2,
                         "startup_config": "iou_l3_base_startup-config.txt",
                         "symbol": ":/symbols/multilayer_switch.svg",
                         "use_default_iou_values": True,
                         "l1_keepalives": False}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_docker_template_create(http_controller):

    params = {"name": "Docker template",
              "compute_id": "local",
              "image": "gns3/endhost:latest",
              "template_type": "docker"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"adapters": 1,
                         "template_type": "docker",
                         "builtin": False,
                         "category": "guest",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_http_path": "/",
                         "console_http_port": 80,
                         "console_resolution": "1024x768",
                         "console_type": "telnet",
                         "default_name_format": "{name}-{0}",
                         "environment": "",
                         "extra_hosts": "",
                         "image": "gns3/endhost:latest",
                         "name": "Docker template",
                         "start_command": "",
                         "symbol": ":/symbols/docker_guest.svg",
                         "custom_adapters": []}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_qemu_template_create(http_controller):

    params = {"name": "Qemu template",
              "compute_id": "local",
              "platform": "i386",
              "hda_disk_image": "IOSvL2-15.2.4.0.55E.qcow2",
              "ram": 512,
              "template_type": "qemu"}

    response = http_controller.post("/templates", params, example=True)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"adapter_type": "e1000",
                         "adapters": 1,
                         "template_type": "qemu",
                         "bios_image": "",
                         "boot_priority": "c",
                         "builtin": False,
                         "category": "guest",
                         "cdrom_image": "",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "cpu_throttling": 0,
                         "cpus": 1,
                         "default_name_format": "{name}-{0}",
                         "first_port_name": "",
                         "hda_disk_image": "IOSvL2-15.2.4.0.55E.qcow2",
                         "hda_disk_interface": "ide",
                         "hdb_disk_image": "",
                         "hdb_disk_interface": "ide",
                         "hdc_disk_image": "",
                         "hdc_disk_interface": "ide",
                         "hdd_disk_image": "",
                         "hdd_disk_interface": "ide",
                         "initrd": "",
                         "kernel_command_line": "",
                         "kernel_image": "",
                         "legacy_networking": False,
                         "linked_clone": True,
                         "mac_address": "",
                         "name": "Qemu template",
                         "on_close": "power_off",
                         "options": "",
                         "platform": "i386",
                         "port_name_format": "Ethernet{0}",
                         "port_segment_size": 0,
                         "process_priority": "normal",
                         "qemu_path": "",
                         "ram": 512,
                         "symbol": ":/symbols/qemu_guest.svg",
                         "usage": "",
                         "custom_adapters": []}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_vmware_template_create(http_controller):

    params = {"name": "VMware template",
              "compute_id": "local",
              "template_type": "vmware",
              "vmx_path": "/path/to/vm.vmx"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"adapter_type": "e1000",
                         "adapters": 1,
                         "template_type": "vmware",
                         "builtin": False,
                         "category": "guest",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "none",
                         "default_name_format": "{name}-{0}",
                         "first_port_name": "",
                         "headless": False,
                         "linked_clone": False,
                         "name": "VMware template",
                         "on_close": "power_off",
                         "port_name_format": "Ethernet{0}",
                         "port_segment_size": 0,
                         "symbol": ":/symbols/vmware_guest.svg",
                         "use_any_adapter": False,
                         "vmx_path": "/path/to/vm.vmx",
                         "custom_adapters": []}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_virtualbox_template_create(http_controller):

    params = {"name": "VirtualBox template",
              "compute_id": "local",
              "template_type": "virtualbox",
              "vmname": "My VirtualBox VM"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"adapter_type": "Intel PRO/1000 MT Desktop (82540EM)",
                         "adapters": 1,
                         "template_type": "virtualbox",
                         "builtin": False,
                         "category": "guest",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "none",
                         "default_name_format": "{name}-{0}",
                         "first_port_name": "",
                         "headless": False,
                         "linked_clone": False,
                         "name": "VirtualBox template",
                         "on_close": "power_off",
                         "port_name_format": "Ethernet{0}",
                         "port_segment_size": 0,
                         "ram": 256,
                         "symbol": ":/symbols/vbox_guest.svg",
                         "use_any_adapter": False,
                         "vmname": "My VirtualBox VM",
                         "custom_adapters": []}

    for item, value in expected_response.items():
        assert response.json.get(item) == value

def test_vpcs_template_create(http_controller):

    params = {"name": "VPCS template",
              "compute_id": "local",
              "template_type": "vpcs"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "vpcs",
                         "base_script_file": "vpcs_base_config.txt",
                         "builtin": False,
                         "category": "guest",
                         "compute_id": "local",
                         "console_auto_start": False,
                         "console_type": "telnet",
                         "default_name_format": "PC{0}",
                         "name": "VPCS template",
                         "symbol": ":/symbols/vpcs_guest.svg"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value

def test_ethernet_switch_template_create(http_controller):

    params = {"name": "Ethernet switch template",
              "compute_id": "local",
              "template_type": "ethernet_switch"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "ethernet_switch",
                         "builtin": False,
                         "category": "switch",
                         "compute_id": "local",
                         "console_type": "none",
                         "default_name_format": "Switch{0}",
                         "name": "Ethernet switch template",
                         "ports_mapping": [{"ethertype": "",
                                            "name": "Ethernet0",
                                            "port_number": 0,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet1",
                                            "port_number": 1,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet2",
                                            "port_number": 2,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet3",
                                            "port_number": 3,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet4",
                                            "port_number": 4,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet5",
                                            "port_number": 5,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet6",
                                            "port_number": 6,
                                            "type": "access",
                                            "vlan": 1
                                            },
                                           {"ethertype": "",
                                            "name": "Ethernet7",
                                            "port_number": 7,
                                            "type": "access",
                                            "vlan": 1
                                            }],
                         "symbol": ":/symbols/ethernet_switch.svg"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_cloud_template_create(http_controller):

    params = {"name": "Cloud template",
              "compute_id": "local",
              "template_type": "cloud"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"template_type": "cloud",
                         "builtin": False,
                         "category": "guest",
                         "compute_id": "local",
                         "default_name_format": "Cloud{0}",
                         "name": "Cloud template",
                         "symbol": ":/symbols/cloud.svg"}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


def test_ethernet_hub_template_create(http_controller):

    params = {"name": "Ethernet hub template",
              "compute_id": "local",
              "template_type": "ethernet_hub"}

    response = http_controller.post("/templates", params)
    assert response.status == 201
    assert response.json["template_id"] is not None

    expected_response = {"ports_mapping": [{"port_number": 0,
                                            "name": "Ethernet0"
                                            },
                                           {"port_number": 1,
                                             "name": "Ethernet1"
                                            },
                                           {"port_number": 2,
                                            "name": "Ethernet2"
                                            },
                                           {"port_number": 3,
                                            "name": "Ethernet3"
                                            },
                                           {"port_number": 4,
                                            "name": "Ethernet4"
                                            },
                                           {"port_number": 5,
                                            "name": "Ethernet5"
                                            },
                                           {"port_number": 6,
                                            "name": "Ethernet6"
                                            },
                                           {"port_number": 7,
                                            "name": "Ethernet7"
                                            }],
                         "compute_id": "local",
                         "name": "Ethernet hub template",
                         "symbol": ":/symbols/hub.svg",
                         "default_name_format": "Hub{0}",
                         "template_type": "ethernet_hub",
                         "category": "switch",
                         "builtin": False}

    for item, value in expected_response.items():
        assert response.json.get(item) == value


@pytest.fixture
def compute(http_controller, async_run):
    compute = MagicMock()
    compute.id = "example.com"
    compute.host = "example.org"
    Controller.instance()._computes = {"example.com": compute}
    return compute


@pytest.fixture
def project(http_controller, async_run):
    return async_run(Controller.instance().add_project(name="Test"))


def test_create_node_from_template(http_controller, controller, project, compute):

    id = str(uuid.uuid4())
    controller.template_manager._templates = {id: Template(id, {
        "template_type": "qemu",
        "category": 0,
        "name": "test",
        "symbol": "guest.svg",
        "default_name_format": "{name}-{0}",
        "compute_id": "example.com"
    })}
    with asyncio_patch("gns3server.controller.project.Project.add_node_from_template", return_value={"name": "test", "node_type": "qemu", "compute_id": "example.com"}) as mock:
        response = http_controller.post("/projects/{}/templates/{}".format(project.id, id), {
            "x": 42,
            "y": 12
        })
    mock.assert_called_with(id, x=42, y=12, compute_id=None)
    assert response.route == "/projects/{project_id}/templates/{template_id}"
    assert response.status == 201



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import aiohttp
from aiohttp.web import WebSocketResponse
from gns3server.web.route import Route
from gns3server.compute.notification_manager import NotificationManager

import logging
log = logging.getLogger(__name__)


async def process_websocket(ws):
    """
    Process ping / pong and close message
    """
    try:
        await ws.receive()
    except aiohttp.WSServerHandshakeError:
        pass


class NotificationHandler:

    @Route.get(
        r"/notifications/ws",
        description="Send notifications using Websockets")
    async def notifications(request, response):
        notifications = NotificationManager.instance()
        ws = WebSocketResponse()
        await ws.prepare(request)

        request.app['websockets'].add(ws)
        asyncio.ensure_future(process_websocket(ws))
        log.info("New client has connected to compute WebSocket")
        try:
            with notifications.queue() as queue:
                while True:
                    notification = await queue.get_json(1)
                    if ws.closed:
                        break
                    await ws.send_str(notification)
        finally:
            log.info("Client has disconnected from compute WebSocket")
            if not ws.closed:
                await ws.close()
            request.app['websockets'].discard(ws)

        return ws



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Interface for Dynamips virtual Cisco 2600 instances module ("c2600")
http://github.com/GNS3/dynamips/blob/master/README.hypervisor#L404
"""

import asyncio
from .router import Router
from ..adapters.c2600_mb_1e import C2600_MB_1E
from ..adapters.c2600_mb_2e import C2600_MB_2E
from ..adapters.c2600_mb_1fe import C2600_MB_1FE
from ..adapters.c2600_mb_2fe import C2600_MB_2FE

import logging
log = logging.getLogger(__name__)


class C2600(Router):

    """
    Dynamips c2600 router.

    :param name: The name of this router
    :param node_id: Node identifier
    :param project: Project instance
    :param manager: Parent VM Manager
    :param dynamips_id: ID to use with Dynamips
    :param console: console port
    :param aux: auxiliary console port
    :param chassis: chassis for this router:
    2610, 2611, 2620, 2621, 2610XM, 2611XM
    2620XM, 2621XM, 2650XM or 2651XM (default = 2610).
    """

    # adapters to insert by default corresponding the
    # chosen chassis.
    integrated_adapters = {"2610": C2600_MB_1E,
                           "2611": C2600_MB_2E,
                           "2620": C2600_MB_1FE,
                           "2621": C2600_MB_2FE,
                           "2610XM": C2600_MB_1FE,
                           "2611XM": C2600_MB_2FE,
                           "2620XM": C2600_MB_1FE,
                           "2621XM": C2600_MB_2FE,
                           "2650XM": C2600_MB_1FE,
                           "2651XM": C2600_MB_2FE}

    def __init__(self, name, node_id, project, manager, dynamips_id, console=None, console_type="telnet", aux=None, chassis="2610"):

        super().__init__(name, node_id, project, manager, dynamips_id, console, console_type, aux, platform="c2600")

        # Set default values for this platform (must be the same as Dynamips)
        self._ram = 64
        self._nvram = 128
        self._disk0 = 0
        self._disk1 = 0
        self._chassis = chassis
        self._iomem = 15  # percentage
        self._clock_divisor = 8
        self._sparsemem = False  # never activate sparsemem for c2600 (unstable)

    def __json__(self):

        c2600_router_info = {"iomem": self._iomem,
                             "chassis": self._chassis,
                             "sparsemem": self._sparsemem}

        router_info = Router.__json__(self)
        router_info.update(c2600_router_info)
        return router_info

    async def create(self):

        await Router.create(self)
        if self._chassis != "2610":
            await self.set_chassis(self._chassis)
        self._setup_chassis()

    def _setup_chassis(self):
        """
        Sets up the router with the corresponding chassis
        (create slots and insert default adapters).
        """

        self._create_slots(2)
        self._slots[0] = self.integrated_adapters[self._chassis]()

    @property
    def chassis(self):
        """
        Returns the chassis.

        :returns: chassis string
        """

        return self._chassis

    async def set_chassis(self, chassis):
        """
        Sets the chassis.

        :param: chassis string:
        2610, 2611, 2620, 2621, 2610XM, 2611XM
        2620XM, 2621XM, 2650XM or 2651XM
        """

        await self._hypervisor.send('c2600 set_chassis "{name}" {chassis}'.format(name=self._name, chassis=chassis))

        log.info('Router "{name}" [{id}]: chassis set to {chassis}'.format(name=self._name,
                                                                           id=self._id,
                                                                           chassis=chassis))
        self._chassis = chassis
        self._setup_chassis()

    @property
    def iomem(self):
        """
        Returns I/O memory size for this router.

        :returns: I/O memory size (integer)
        """

        return self._iomem

    async def set_iomem(self, iomem):
        """
        Sets I/O memory size for this router.

        :param iomem: I/O memory size
        """

        await self._hypervisor.send('c2600 set_iomem "{name}" {size}'.format(name=self._name, size=iomem))

        log.info('Router "{name}" [{id}]: I/O memory updated from {old_iomem}% to {new_iomem}%'.format(name=self._name,
                                                                                                       id=self._id,
                                                                                                       old_iomem=self._iomem,
                                                                                                       new_iomem=iomem))
        self._iomem = iomem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import aiohttp
import pytest
import sys
import uuid
from unittest.mock import patch

from gns3server.compute.port_manager import PortManager
from gns3server.compute.project import Project


def test_reserve_tcp_port():
    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    pm.reserve_tcp_port(2001, project)
    with patch("gns3server.compute.project.Project.emit") as mock_emit:
        port = pm.reserve_tcp_port(2001, project)
        assert port != 2001


def test_reserve_tcp_port_outside_range():
    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    with patch("gns3server.compute.project.Project.emit") as mock_emit:
        port = pm.reserve_tcp_port(80, project)
        assert port != 80


def test_reserve_tcp_port_already_used_by_another_program():
    """
    This test simulate a scenario where the port is already taken
    by another programm on the server
    """

    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    with patch("gns3server.compute.port_manager.PortManager._check_port") as mock_check:

        def execute_mock(host, port, *args):
            if port == 2001:
                raise OSError("Port is already used")
            else:
                return True

        mock_check.side_effect = execute_mock

        with patch("gns3server.compute.project.Project.emit") as mock_emit:
            port = pm.reserve_tcp_port(2001, project)
            assert port != 2001


def test_reserve_tcp_port_already_used():
    """
    This test simulate a scenario where the port is already taken
    by another programm on the server
    """

    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    with patch("gns3server.compute.port_manager.PortManager._check_port") as mock_check:

        def execute_mock(host, port, *args):
            if port == 2001:
                raise OSError("Port is already used")
            else:
                return True

        mock_check.side_effect = execute_mock

        with patch("gns3server.compute.project.Project.emit") as mock_emit:
            port = pm.reserve_tcp_port(2001, project)
            assert port != 2001


def test_reserve_udp_port():
    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    pm.reserve_udp_port(10000, project)
    with pytest.raises(aiohttp.web.HTTPConflict):
        pm.reserve_udp_port(10000, project)


def test_reserve_udp_port_outside_range():
    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    with pytest.raises(aiohttp.web.HTTPConflict):
        pm.reserve_udp_port(80, project)


def test_release_udp_port():
    pm = PortManager()
    project = Project(project_id=str(uuid.uuid4()))
    pm.reserve_udp_port(10000, project)
    pm.release_udp_port(10000, project)
    pm.reserve_udp_port(10000, project)


def test_find_unused_port():
    p = PortManager().find_unused_port(1000, 10000)
    assert p is not None


def test_find_unused_port_invalid_range():
    with pytest.raises(aiohttp.web.HTTPConflict):
        p = PortManager().find_unused_port(10000, 1000)


def test_set_console_host(config):
    """
    If allow remote connection we need to bind console host
    to 0.0.0.0
    """
    p = PortManager()
    config.set_section_config("Server", {"allow_remote_console": False})
    p.console_host = "10.42.1.42"
    assert p.console_host == "10.42.1.42"
    p = PortManager()
    config.set_section_config("Server", {"allow_remote_console": True})
    p.console_host = "10.42.1.42"
    assert p.console_host == "0.0.0.0"



#!/usr/bin/env python
#
# Copyright (C) 2016 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import json
import uuid
import socket
import shutil
import aiohttp

from ..config import Config
from .project import Project
from .template import Template
from .appliance import Appliance
from .appliance_manager import ApplianceManager
from .template_manager import TemplateManager
from .compute import Compute, ComputeError
from .notification import Notification
from .symbols import Symbols
from ..version import __version__
from .topology import load_topology
from .gns3vm import GNS3VM
from ..utils.get_resource import get_resource
from .gns3vm.gns3_vm_error import GNS3VMError

import logging
log = logging.getLogger(__name__)


class Controller:
    """
    The controller is responsible to manage one or more computes.
    """

    def __init__(self):
        self._computes = {}
        self._projects = {}
        self._notification = Notification(self)
        self.gns3vm = GNS3VM(self)
        self.symbols = Symbols()
        self._appliance_manager = ApplianceManager()
        self._template_manager = TemplateManager()
        self._iou_license_settings = {"iourc_content": "",
                                      "license_check": True}
        self._config_loaded = False
        self._config_file = Config.instance().controller_config
        log.info("Load controller configuration file {}".format(self._config_file))

    async def start(self):

        log.info("Controller is starting")
        self.load_base_files()
        server_config = Config.instance().get_section_config("Server")
        Config.instance().listen_for_config_changes(self._update_config)
        host = server_config.get("host", "localhost")
        port = server_config.getint("port", 3080)

        # clients will use the IP they use to connect to
        # the controller if console_host is 0.0.0.0
        console_host = host
        if host == "0.0.0.0":
            host = "127.0.0.1"

        name = socket.gethostname()
        if name == "gns3vm":
            name = "Main server"

        computes = await self._load_controller_settings()
        try:
            self._local_server = await self.add_compute(compute_id="local",
                                                        name=name,
                                                        protocol=server_config.get("protocol", "http"),
                                                        host=host,
                                                        console_host=console_host,
                                                        port=port,
                                                        user=server_config.get("user", ""),
                                                        password=server_config.get("password", ""),
                                                        force=True)
        except aiohttp.web.HTTPConflict:
            log.fatal("Cannot access to the local server, make sure something else is not running on the TCP port {}".format(port))
            sys.exit(1)
        for c in computes:
            try:
                await self.add_compute(**c)
            except (aiohttp.web.HTTPError, KeyError):
                pass  # Skip not available servers at loading
        await self.load_projects()
        try:
            await self.gns3vm.auto_start_vm()
        except GNS3VMError as e:
            log.warning(str(e))
        await self._project_auto_open()

    def _update_config(self):
        """
        Call this when the server configuration file changes.
        """

        if self._local_server:
            server_config = Config.instance().get_section_config("Server")
            self._local_server.user = server_config.get("user")
            self._local_server.password = server_config.get("password")

    async def stop(self):

        log.info("Controller is Stopping")
        for project in self._projects.values():
            await project.close()
        for compute in self._computes.values():
            try:
                await compute.close()
            # We don't care if a compute is down at this step
            except (ComputeError, aiohttp.web.HTTPError, OSError):
                pass
        await self.gns3vm.exit_vm()
        #self.save()
        self._computes = {}
        self._projects = {}

    def save(self):
        """
        Save the controller configuration on disk
        """

        if self._config_loaded is False:
            return

        controller_settings = {"computes": [],
                               "templates": [],
                               "gns3vm": self.gns3vm.__json__(),
                               "iou_license": self._iou_license_settings,
                               "appliances_etag": self._appliance_manager.appliances_etag,
                               "version": __version__}

        for template in self._template_manager.templates.values():
            if not template.builtin:
                controller_settings["templates"].append(template.__json__())

        for compute in self._computes.values():
            if compute.id != "local" and compute.id != "vm":
                controller_settings["computes"].append({"host": compute.host,
                                                        "name": compute.name,
                                                        "port": compute.port,
                                                        "protocol": compute.protocol,
                                                        "user": compute.user,
                                                        "password": compute.password,
                                                        "compute_id": compute.id})

        try:
            os.makedirs(os.path.dirname(self._config_file), exist_ok=True)
            with open(self._config_file, 'w+') as f:
                json.dump(controller_settings, f, indent=4)
        except OSError as e:
            log.error("Cannot write controller configuration file '{}': {}".format(self._config_file, e))

    async def _load_controller_settings(self):
        """
        Reload the controller configuration from disk
        """

        try:
            if not os.path.exists(self._config_file):
                self._config_loaded = True
                self.save()
            with open(self._config_file) as f:
                controller_settings = json.load(f)
        except (OSError, ValueError) as e:
            log.critical("Cannot load configuration file '{}': {}".format(self._config_file, e))
            return []

        # load GNS3 VM settings
        if "gns3vm" in controller_settings:
            self.gns3vm.settings = controller_settings["gns3vm"]

        # load the IOU license settings
        if "iou_license" in controller_settings:
            self._iou_license_settings = controller_settings["iou_license"]

        self._appliance_manager.appliances_etag = controller_settings.get("appliances_etag")
        self._appliance_manager.load_appliances()
        self._template_manager.load_templates(controller_settings.get("templates"))
        self._config_loaded = True
        return controller_settings.get("computes", [])

    async def load_projects(self):
        """
        Preload the list of projects from disk
        """

        server_config = Config.instance().get_section_config("Server")
        projects_path = os.path.expanduser(server_config.get("projects_path", "~/GNS3/projects"))
        os.makedirs(projects_path, exist_ok=True)
        try:
            for project_path in os.listdir(projects_path):
                project_dir = os.path.join(projects_path, project_path)
                if os.path.isdir(project_dir):
                    for file in os.listdir(project_dir):
                        if file.endswith(".gns3"):
                            try:
                                await self.load_project(os.path.join(project_dir, file), load=False)
                            except (aiohttp.web.HTTPConflict, NotImplementedError):
                                pass  # Skip not compatible projects
        except OSError as e:
            log.error(str(e))

    def load_base_files(self):
        """
        At startup we copy base file to the user location to allow
        them to customize it
        """

        dst_path = self.configs_path()
        src_path = get_resource('configs')
        try:
            for file in os.listdir(src_path):
                if not os.path.exists(os.path.join(dst_path, file)):
                    shutil.copy(os.path.join(src_path, file), os.path.join(dst_path, file))
        except OSError:
            pass

    def images_path(self):
        """
        Get the image storage directory
        """

        server_config = Config.instance().get_section_config("Server")
        images_path = os.path.expanduser(server_config.get("images_path", "~/GNS3/projects"))
        os.makedirs(images_path, exist_ok=True)
        return images_path

    def configs_path(self):
        """
        Get the configs storage directory
        """

        server_config = Config.instance().get_section_config("Server")
        images_path = os.path.expanduser(server_config.get("configs_path", "~/GNS3/projects"))
        os.makedirs(images_path, exist_ok=True)
        return images_path

    async def add_compute(self, compute_id=None, name=None, force=False, connect=True, **kwargs):
        """
        Add a server to the dictionary of computes controlled by this controller

        :param compute_id: Compute identifier
        :param name: Compute name
        :param force: True skip security check
        :param connect: True connect to the compute immediately
        :param kwargs: See the documentation of Compute
        """

        if compute_id not in self._computes:

            # We disallow to create from the outside the local and VM server
            if (compute_id == 'local' or compute_id == 'vm') and not force:
                return None

            # It seem we have error with a gns3vm imported as a remote server and conflict
            # with GNS3 VM settings. That's why we ignore server name gns3vm
            if name == 'gns3vm':
                return None

            for compute in self._computes.values():
                if name and compute.name == name and not force:
                    raise aiohttp.web.HTTPConflict(text='Compute name "{}" already exists'.format(name))

            compute = Compute(compute_id=compute_id, controller=self, name=name, **kwargs)
            self._computes[compute.id] = compute
            self.save()
            if connect:
                await compute.connect()
            self.notification.controller_emit("compute.created", compute.__json__())
            return compute
        else:
            if connect:
                await self._computes[compute_id].connect()
            self.notification.controller_emit("compute.updated", self._computes[compute_id].__json__())
            return self._computes[compute_id]

    async def close_compute_projects(self, compute):
        """
        Close projects running on a compute
        """

        for project in self._projects.values():
            if compute in project.computes:
                await project.close()

    def compute_has_open_project(self, compute):
        """
        Check is compute has a project opened.

        :returns: True if a project is open
        """

        for project in self._projects.values():
            if compute in project.computes and project.status == "opened":
                return True
        return False

    async def delete_compute(self, compute_id):
        """
        Delete a compute node. Project using this compute will be close

        :param compute_id: Compute identifier
        """

        try:
            compute = self.get_compute(compute_id)
        except aiohttp.web.HTTPNotFound:
            return
        await self.close_compute_projects(compute)
        await compute.close()
        del self._computes[compute_id]
        self.save()
        self.notification.controller_emit("compute.deleted", compute.__json__())

    @property
    def notification(self):
        """
        The notification system
        """

        return self._notification

    @property
    def computes(self):
        """
        :returns: The dictionary of computes managed by this controller
        """

        return self._computes

    def get_compute(self, compute_id):
        """
        Returns a compute or raise a 404 error.
        """

        try:
            return self._computes[compute_id]
        except KeyError:
            if compute_id == "vm":
                raise aiohttp.web.HTTPNotFound(text="Cannot use a node on the GNS3 VM server with the GNS3 VM not configured")
            raise aiohttp.web.HTTPNotFound(text="Compute ID {} doesn't exist".format(compute_id))

    def has_compute(self, compute_id):
        """
        Return True if the compute exist in the controller
        """

        return compute_id in self._computes

    async def add_project(self, project_id=None, name=None, path=None, **kwargs):
        """
        Creates a project or returns an existing project

        :param project_id: Project ID
        :param name: Project name
        :param kwargs: See the documentation of Project
        """

        if project_id not in self._projects:
            for project in self._projects.values():
                if name and project.name == name:
                    if path and path == project.path:
                        raise aiohttp.web.HTTPConflict(text='Project "{}" already exists in location "{}"'.format(name, path))
                    else:
                        raise aiohttp.web.HTTPConflict(text='Project "{}" already exists'.format(name))
            project = Project(project_id=project_id, controller=self, name=name, path=path, **kwargs)
            self._projects[project.id] = project
            return self._projects[project.id]
        return self._projects[project_id]

    def get_project(self, project_id):
        """
        Returns a project or raise a 404 error.
        """

        try:
            return self._projects[project_id]
        except KeyError:
            raise aiohttp.web.HTTPNotFound(text="Project ID {} doesn't exist".format(project_id))

    async def get_loaded_project(self, project_id):
        """
        Returns a project or raise a 404 error.

        If project is not finished to load wait for it
        """

        project = self.get_project(project_id)
        await project.wait_loaded()
        return project

    def remove_project(self, project):

        if project.id in self._projects:
            del self._projects[project.id]

    async def load_project(self, path, load=True):
        """
        Load a project from a .gns3

        :param path: Path of the .gns3
        :param load: Load the topology
        """

        topo_data = load_topology(path)
        topo_data.pop("topology")
        topo_data.pop("version")
        topo_data.pop("revision")
        topo_data.pop("type")

        if topo_data["project_id"] in self._projects:
            project = self._projects[topo_data["project_id"]]
        else:
            project = await self.add_project(path=os.path.dirname(path), status="closed", filename=os.path.basename(path), **topo_data)
        if load or project.auto_open:
            await project.open()
        return project

    async def _project_auto_open(self):
        """
        Auto open the project with auto open enable
        """

        for project in self._projects.values():
            if project.auto_open:
                await project.open()

    def get_free_project_name(self, base_name):
        """
        Generate a free project name base on the base name
        """

        names = [p.name for p in self._projects.values()]
        if base_name not in names:
            return base_name
        i = 1

        projects_path = self.projects_directory()

        while True:
            new_name = "{}-{}".format(base_name, i)
            if new_name not in names and not os.path.exists(os.path.join(projects_path, new_name)):
                break
            i += 1
            if i > 1000000:
                raise aiohttp.web.HTTPConflict(text="A project name could not be allocated (node limit reached?)")
        return new_name

    @property
    def projects(self):
        """
        :returns: The dictionary of projects managed by GNS3
        """

        return self._projects

    @property
    def appliance_manager(self):
        """
        :returns: Appliance Manager instance
        """

        return self._appliance_manager

    @property
    def template_manager(self):
        """
        :returns: Template Manager instance
        """

        return self._template_manager

    @property
    def iou_license(self):
        """
        :returns: The dictionary of IOU license settings
        """

        return self._iou_license_settings

    def projects_directory(self):

        server_config = Config.instance().get_section_config("Server")
        return os.path.expanduser(server_config.get("projects_path", "~/GNS3/projects"))

    @staticmethod
    def instance():
        """
        Singleton to return only on instance of Controller.

        :returns: instance of Controller
        """

        if not hasattr(Controller, '_instance') or Controller._instance is None:
            Controller._instance = Controller()
        return Controller._instance

    async def autoidlepc(self, compute_id, platform, image, ram):
        """
        Compute and IDLE PC value for an image

        :param compute_id: ID of the compute where the idlepc operation need to run
        :param platform: Platform type
        :param image: Image to use
        :param ram: amount of RAM to use
        """

        compute = self.get_compute(compute_id)
        for project in list(self._projects.values()):
            if project.name == "AUTOIDLEPC":
                await project.delete()
                self.remove_project(project)
        project = await self.add_project(name="AUTOIDLEPC")
        node = await project.add_node(compute, "AUTOIDLEPC", str(uuid.uuid4()), node_type="dynamips", platform=platform, image=image, ram=ram)
        res = await node.dynamips_auto_idlepc()
        await project.delete()
        self.remove_project(project)
        return res

    async def compute_ports(self, compute_id):
        """
        Get the ports used by a compute.

        :param compute_id: ID of the compute
        """

        compute = self.get_compute(compute_id)
        response = await compute.get("/network/ports")
        return response.json



#!/usr/bin/env python
#
# Copyright (C) 2016 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import asyncio
import struct


class Qcow2Error(Exception):
    pass


class Qcow2:
    """
    Allows to parse a Qcow2 file
    """

    def __init__(self, path):

        self._path = path
        self._reload()

    def _reload(self):
        # Each QCOW2 file begins with a header, in big endian format, as follows:
        #
        # typedef struct QCowHeader {
        #     uint32_t magic;
        #     uint32_t version;
        #
        #     uint64_t backing_file_offset;
        #     uint32_t backing_file_size;
        #
        #     uint32_t cluster_bits;
        #     uint64_t size; /* in bytes */
        #     uint32_t crypt_method;
        #
        #     uint32_t l1_size;
        #     uint64_t l1_table_offset;
        #
        #     uint64_t refcount_table_offset;
        #     uint32_t refcount_table_clusters;
        #
        #     uint32_t nb_snapshots;
        #     uint64_t snapshots_offset;
        # } QCowHeader;

        struct_format = ">IIQi"
        with open(self._path, 'rb') as f:
            content = f.read(struct.calcsize(struct_format))
            try:
                self.magic, self.version, self.backing_file_offset, self.backing_file_size = struct.unpack_from(struct_format, content)

            except struct.error:
                raise Qcow2Error("Invalid file header for {}".format(self._path))

        if self.magic != 1363560955:  # The first 4 bytes contain the characters 'Q', 'F', 'I' followed by 0xfb.
            raise Qcow2Error("Invalid magic for {}".format(self._path))

    @property
    def backing_file(self):
        """
        When using linked clone this will return the path to the base image

        :returns: None if it's not a linked clone, the path otherwise
        """

        with open(self._path, 'rb') as f:
            f.seek(self.backing_file_offset)
            content = f.read(self.backing_file_size)

        path = content.decode()
        if len(path) == 0:
            return None
        return path

    async def rebase(self, qemu_img, base_image):
        """
        Rebase a linked clone in order to use the correct disk

        :param qemu_img: Path to the qemu-img binary
        :param base_image: Path to the base image
        """

        if not os.path.exists(base_image):
            raise FileNotFoundError(base_image)
        command = [qemu_img, "rebase", "-u", "-b", base_image, self._path]
        process = await asyncio.create_subprocess_exec(*command)
        retcode = await process.wait()
        if retcode != 0:
            raise Qcow2Error("Could not rebase the image")
        self._reload()
